"""Attachment content types for XMTP."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, cast
from urllib.parse import urlparse

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _FfiAttachment(Protocol):
    def __init__(self, *, filename: str | None, mime_type: str, content: bytes) -> None: ...

    filename: str | None
    mime_type: str
    content: bytes


class _FfiRemoteAttachment(Protocol):
    def __init__(
        self,
        *,
        url: str,
        content_digest: str,
        secret: bytes,
        salt: bytes,
        nonce: bytes,
        scheme: str,
        content_length: int,
        filename: str | None,
    ) -> None: ...

    url: str
    content_digest: str
    secret: bytes
    salt: bytes
    nonce: bytes
    scheme: str
    content_length: int
    filename: str | None


class _Bindings(Protocol):
    FfiAttachment: type[_FfiAttachment]
    FfiRemoteAttachment: type[_FfiRemoteAttachment]

    def encode_attachment(self, attachment: _FfiAttachment) -> bytes: ...

    def decode_attachment(self, data: bytes) -> _FfiAttachment: ...

    def encode_remote_attachment(self, attachment: _FfiRemoteAttachment) -> bytes: ...

    def decode_remote_attachment(self, data: bytes) -> _FfiRemoteAttachment: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeAttachment = ContentTypeId(
    authority_id="xmtp.org",
    type_id="attachment",
    version_major=1,
    version_minor=0,
)

ContentTypeRemoteAttachment = ContentTypeId(
    authority_id="xmtp.org",
    type_id="remoteStaticAttachment",
    version_major=1,
    version_minor=0,
)


@dataclass(slots=True)
class Attachment:
    """Attachment payload."""

    filename: str | None
    mime_type: str
    data: bytes


@dataclass(slots=True)
class RemoteAttachment:
    """Remote attachment metadata."""

    url: str
    content_digest: str
    salt: bytes
    nonce: bytes
    secret: bytes
    scheme: str
    content_length: int
    filename: str | None


class AttachmentCodec(ContentCodec[Attachment]):
    """Codec for attachment payloads."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeAttachment

    def encode(self, content: Attachment, registry: CodecRegistry | None = None) -> EncodedContent:
        attachment = _bindings().FfiAttachment(
            filename=content.filename,
            mime_type=content.mime_type,
            content=content.data,
        )
        encoded = _bindings().encode_attachment(attachment)
        parameters = {"mimeType": content.mime_type}
        if content.filename:
            parameters["filename"] = content.filename
        return EncodedContent(type_id=self.content_type, parameters=parameters, content=encoded)

    def decode(self, content: EncodedContent, registry: CodecRegistry | None = None) -> Attachment:
        decoded = _bindings().decode_attachment(content.content)
        return Attachment(
            filename=decoded.filename,
            mime_type=decoded.mime_type,
            data=decoded.content,
        )

    def fallback(self, content: Attachment) -> str | None:
        filename = content.filename or "attachment"
        return f"Can't display \"{filename}\". This app doesn't support attachments."

    def should_push(self, content: Attachment) -> bool:
        return True


class RemoteAttachmentCodec(ContentCodec[RemoteAttachment]):
    """Codec for remote attachment metadata."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeRemoteAttachment

    def encode(
        self, content: RemoteAttachment, registry: CodecRegistry | None = None
    ) -> EncodedContent:
        parsed = urlparse(content.url)
        if parsed.scheme.lower() != "https":
            raise ValueError("Remote attachment URL must use https scheme")
        remote_attachment = _bindings().FfiRemoteAttachment(
            url=content.url,
            content_digest=content.content_digest,
            secret=content.secret,
            salt=content.salt,
            nonce=content.nonce,
            scheme=content.scheme,
            content_length=content.content_length,
            filename=content.filename,
        )
        encoded = _bindings().encode_remote_attachment(remote_attachment)
        parameters = {
            "contentDigest": content.content_digest,
            "salt": content.salt.hex(),
            "nonce": content.nonce.hex(),
            "secret": content.secret.hex(),
            "scheme": content.scheme,
            "contentLength": str(content.content_length),
        }
        if content.filename:
            parameters["filename"] = content.filename
        return EncodedContent(type_id=self.content_type, parameters=parameters, content=encoded)

    def decode(
        self, content: EncodedContent, registry: CodecRegistry | None = None
    ) -> RemoteAttachment:
        decoded = _bindings().decode_remote_attachment(content.content)
        return RemoteAttachment(
            url=decoded.url,
            content_digest=decoded.content_digest,
            secret=decoded.secret,
            salt=decoded.salt,
            nonce=decoded.nonce,
            scheme=decoded.scheme,
            content_length=decoded.content_length,
            filename=decoded.filename,
        )

    def fallback(self, content: RemoteAttachment) -> str | None:
        filename = content.filename or "attachment"
        return f"Can't display \"{filename}\". This app doesn't support attachments."

    def should_push(self, content: RemoteAttachment) -> bool:
        return True


__all__ = [
    "Attachment",
    "AttachmentCodec",
    "ContentTypeAttachment",
    "ContentTypeRemoteAttachment",
    "RemoteAttachment",
    "RemoteAttachmentCodec",
]



"""User helpers for XMTP agents."""

from __future__ import annotations

from dataclasses import dataclass

from eth_account import Account
from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.signers import EoaSigner
from xmtp.signers import create_signer as _create_signer
from xmtp.utils import is_hex_string


@dataclass(slots=True)
class User:
    """User wallet wrapper."""

    private_key: str
    address: str


def _normalize_key(private_key: str) -> str:
    if not is_hex_string(private_key):
        raise ValueError("Private key must be a hex string")
    if not private_key.startswith("0x"):
        return f"0x{private_key}"
    return private_key


def create_user(private_key: str | None = None) -> User:
    """Create a user from an optional private key."""

    if private_key is None:
        account = Account.create()
        key_hex = account.key.hex()
        key_hex = f"0x{key_hex}"
        return User(private_key=key_hex, address=account.address)

    key = _normalize_key(private_key)
    account = Account.from_key(key)
    return User(private_key=key, address=account.address)


def create_signer(user_or_key: User | str) -> EoaSigner:
    """Create an EOA signer from a user or private key."""

    key = user_or_key.private_key if isinstance(user_or_key, User) else user_or_key
    return _create_signer(_normalize_key(key))


def create_identifier(user: User) -> Identifier:
    """Create an XMTP identifier for a user."""

    return Identifier(kind=IdentifierKind.ETHEREUM, value=user.address.lower())


__all__ = ["User", "create_user", "create_signer", "create_identifier"]



"""Middleware interfaces for the agent SDK."""

from __future__ import annotations

import asyncio
import time
from collections.abc import Awaitable, Callable

from xmtp_agent.context import ClientContext, ConversationContext, MessageContext

Handler = Callable[[MessageContext], Awaitable[None]]
ConversationHandler = Callable[[ConversationContext], Awaitable[None]]
LifecycleHandler = Callable[[ClientContext], Awaitable[None]]
NextHandler = Callable[[], Awaitable[None]]
ErrorNextHandler = Callable[[Exception | None], Awaitable[None]]
Middleware = Callable[[MessageContext, NextHandler], Awaitable[None]]
ErrorMiddleware = Callable[
    [Exception, MessageContext | ClientContext, ErrorNextHandler],
    Awaitable[None],
]


def backoff_reconnect(
    *,
    initial_delay: float = 0.5,
    max_delay: float = 30.0,
    multiplier: float = 2.0,
    reset_after: float = 60.0,
) -> ErrorMiddleware:
    """Return error middleware that handles stream errors with backoff."""

    delay = initial_delay
    last_error_at: float | None = None

    async def _middleware(
        error: Exception,
        context: MessageContext | ClientContext,
        next_handler: ErrorNextHandler,
    ) -> None:
        nonlocal delay, last_error_at
        from xmtp_agent.errors import AgentStreamingError

        if not isinstance(error, AgentStreamingError):
            await next_handler(error)
            return

        now = time.monotonic()
        if last_error_at is not None and (now - last_error_at) > reset_after:
            delay = initial_delay

        await asyncio.sleep(delay)
        last_error_at = now
        delay = min(max_delay, delay * multiplier)
        await next_handler(None)

    return _middleware


__all__ = [
    "ErrorMiddleware",
    "ErrorNextHandler",
    "Handler",
    "LifecycleHandler",
    "Middleware",
    "NextHandler",
    "ConversationHandler",
    "backoff_reconnect",
]



"""Environment variable helpers for XMTP."""

from __future__ import annotations

import os
from dataclasses import replace
from typing import cast

from xmtp.signers import Signer, create_signer
from xmtp.types import ClientOptions, LogLevel, XmtpEnv
from xmtp.utils import is_hex_string


def _is_truthy(value: str | None) -> bool:
    if value is None:
        return False
    return value.strip().lower() not in {"", "0", "false", "no", "off"}


def _parse_log_level(value: str | None) -> LogLevel | None:
    if value is None:
        return None
    try:
        return LogLevel(value)
    except ValueError:
        return None


def load_client_options_from_env(options: ClientOptions | None = None) -> ClientOptions:
    """Load client options overrides from environment variables."""

    base = options or ClientOptions()
    updated = replace(base)

    env_value = os.getenv("XMTP_ENV")
    if env_value in {"local", "dev", "production"}:
        updated.env = cast(XmtpEnv, env_value)

    api_url = os.getenv("XMTP_API_URL")
    if api_url:
        updated.api_url = api_url

    history_sync_url = os.getenv("XMTP_HISTORY_SYNC_URL")
    if history_sync_url is not None:
        normalized = history_sync_url.strip().lower()
        if normalized in {"", "none", "disable", "disabled", "off"}:
            updated.disable_history_sync = True
            updated.history_sync_url = None
        else:
            updated.history_sync_url = history_sync_url
            updated.disable_history_sync = False

    gateway_host = os.getenv("XMTP_GATEWAY_HOST")
    if gateway_host:
        updated.gateway_host = gateway_host

    rust_log = os.getenv("XMTP_RUST_LOG")
    if rust_log is not None:
        updated.rust_log = rust_log

    db_encryption_key = os.getenv("XMTP_DB_ENCRYPTION_KEY")
    if db_encryption_key:
        updated.db_encryption_key = db_encryption_key

    db_directory = os.getenv("XMTP_DB_DIRECTORY")
    if db_directory:
        os.makedirs(db_directory, exist_ok=True, mode=0o700)

        def db_path(inbox_id: str) -> str:
            path = os.path.join(db_directory, f"xmtp-{inbox_id}.db3")
            print(f'Saving local database to "{path}"')
            return path

        updated.db_path = db_path

    if _is_truthy(os.getenv("XMTP_FORCE_DEBUG")):
        updated.debug_events_enabled = True
        updated.structured_logging = True
        level = _parse_log_level(os.getenv("XMTP_FORCE_DEBUG_LEVEL"))
        updated.logging_level = level or LogLevel.WARN

    if _is_truthy(os.getenv("XMTP_DISABLE_HISTORY_SYNC")):
        updated.disable_history_sync = True
        updated.history_sync_url = None

    return updated


def apply_rust_log_from_options(options: ClientOptions) -> None:
    """Apply Rust log filtering based on client options."""

    if options.rust_log is not None:
        os.environ["RUST_LOG"] = options.rust_log
    else:
        os.environ.setdefault("RUST_LOG", "off")


def load_signer_from_env() -> Signer:
    """Create an EOA signer from XMTP_WALLET_KEY."""

    wallet_key = os.getenv("XMTP_WALLET_KEY")
    if wallet_key is None:
        raise ValueError("XMTP_WALLET_KEY is not set")
    if not is_hex_string(wallet_key):
        raise ValueError("XMTP_WALLET_KEY must be a hex string")
    if not wallet_key.startswith("0x"):
        wallet_key = f"0x{wallet_key}"
    return create_signer(wallet_key)


__all__ = [
    "apply_rust_log_from_options",
    "load_client_options_from_env",
    "load_signer_from_env",
]



"""Conversation management."""

from __future__ import annotations

import asyncio
import builtins
import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast

from xmtp.async_stream import AsyncStream
from xmtp.bindings import NativeBindings
from xmtp.conversation import Conversation, Dm, Group
from xmtp.errors import ClientNotInitializedError
from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.messages import DecodedMessage

if TYPE_CHECKING:
    from xmtp.client import Client


@dataclass(slots=True)
class ListConversationsOptions:
    """Options for listing conversations."""

    created_after_ns: int | None = None
    created_before_ns: int | None = None
    last_activity_before_ns: int | None = None
    last_activity_after_ns: int | None = None
    order_by: NativeBindings.FfiGroupQueryOrderBy | None = None
    limit: int | None = None
    consent_states: list[NativeBindings.FfiConsentState] | None = None
    include_duplicate_dms: bool = False

    def to_ffi(self) -> NativeBindings.FfiListConversationsOptions:
        return NativeBindings.FfiListConversationsOptions(
            created_after_ns=self.created_after_ns,
            created_before_ns=self.created_before_ns,
            last_activity_before_ns=self.last_activity_before_ns,
            last_activity_after_ns=self.last_activity_after_ns,
            order_by=self.order_by,
            limit=self.limit,
            consent_states=self.consent_states,
            include_duplicate_dms=self.include_duplicate_dms,
        )


def _identifier_to_ffi(identifier: Identifier) -> NativeBindings.FfiIdentifier:
    kind = {
        IdentifierKind.ETHEREUM: NativeBindings.FfiIdentifierKind.ETHEREUM,
        IdentifierKind.PASSKEY: NativeBindings.FfiIdentifierKind.PASSKEY,
    }[identifier.kind]
    return NativeBindings.FfiIdentifier(identifier=identifier.value, identifier_kind=kind)


def _default_dm_options() -> NativeBindings.FfiCreateDmOptions:
    return NativeBindings.FfiCreateDmOptions(message_disappearing_settings=None)


def _default_group_options() -> NativeBindings.FfiCreateGroupOptions:
    return NativeBindings.FfiCreateGroupOptions(
        permissions=None,
        group_name=None,
        group_image_url_square=None,
        group_description=None,
        custom_permission_policy_set=None,
        message_disappearing_settings=None,
        app_data=None,
    )


def _default_list_options() -> NativeBindings.FfiListConversationsOptions:
    return ListConversationsOptions().to_ffi()


class Conversations:
    """Manage XMTP conversations."""

    def __init__(
        self,
        client: Client,
        ffi_conversations: NativeBindings.FfiConversations | None,
    ) -> None:
        self._client = client
        self._ffi = ffi_conversations

    def _ensure(self) -> NativeBindings.FfiConversations:
        if self._ffi is None:
            raise ClientNotInitializedError()
        return self._ffi

    async def new_dm(self, address: str) -> Dm:
        """Create a new direct message conversation by address."""

        identifier = Identifier(kind=IdentifierKind.ETHEREUM, value=address)
        return await self.new_dm_with_identifier(identifier)

    async def new_dm_with_identifier(self, identifier: Identifier) -> Dm:
        """Create a new direct message using an identifier."""

        ffi = self._ensure()
        options = _default_dm_options()

        if hasattr(ffi, "find_or_create_dm_by_identity"):
            try:
                group = await ffi.find_or_create_dm_by_identity(
                    _identifier_to_ffi(identifier),
                    options,
                )
                return Dm(self._client, group)
            except TypeError:
                try:
                    group = await ffi.find_or_create_dm_by_identity(identifier.value, options)
                    return Dm(self._client, group)
                except TypeError:
                    pass

        inbox_id = await self._client.get_inbox_id_by_identifier(identifier)
        if inbox_id is None:
            raise ValueError(f'No inbox id found for identifier "{identifier.value}"')

        if hasattr(ffi, "find_or_create_dm_by_inbox_id"):
            try:
                group = await ffi.find_or_create_dm_by_inbox_id(inbox_id, options)
                return Dm(self._client, group)
            except TypeError:
                pass

        try:
            group = await ffi.find_or_create_dm(_identifier_to_ffi(identifier), options)
        except TypeError:
            group = await cast(Any, ffi).find_or_create_dm(inbox_id, options)
        return Dm(self._client, group)

    async def new_group(self, members: list[str]) -> Group:
        """Create a new group conversation by member addresses."""

        identifiers = [Identifier(kind=IdentifierKind.ETHEREUM, value=member) for member in members]
        return await self.new_group_with_identifiers(identifiers)

    async def new_group_with_identifiers(self, identifiers: list[Identifier]) -> Group:
        """Create a new group conversation using identifiers."""

        group = await self._ensure().create_group(
            [_identifier_to_ffi(identifier) for identifier in identifiers],
            _default_group_options(),
        )
        return Group(self._client, group)

    async def list(
        self,
        options: ListConversationsOptions | None = None,
    ) -> builtins.list[Conversation]:
        """List all conversations."""

        items = self._ensure().list((options or ListConversationsOptions()).to_ffi())
        return [self._wrap_conversation(item.conversation()) for item in items]

    async def list_dms(
        self,
        options: ListConversationsOptions | None = None,
    ) -> builtins.list[Dm]:
        """List direct message conversations."""

        items = self._ensure().list_dms((options or ListConversationsOptions()).to_ffi())
        return [Dm(self._client, item.conversation()) for item in items]

    async def list_groups(
        self,
        options: ListConversationsOptions | None = None,
    ) -> builtins.list[Group]:
        """List group conversations."""

        items = self._ensure().list_groups((options or ListConversationsOptions()).to_ffi())
        return [Group(self._client, item.conversation()) for item in items]

    async def get_conversation_by_id(self, conversation_id: bytes) -> Conversation | None:
        """Return a conversation by identifier."""

        try:
            if self._client._client is None:
                raise ClientNotInitializedError()
            convo = self._client._client.conversation(conversation_id)
        except Exception:
            return None
        return self._wrap_conversation(convo)

    def stream(self) -> AsyncStream[Conversation | NativeBindings.FfiSubscribeError]:
        """Stream new conversations."""

        queue: asyncio.Queue[object] = asyncio.Queue()
        loop = asyncio.get_event_loop()

        class Callback(NativeBindings.FfiConversationCallback):
            def on_conversation(
                self,
                conversation: NativeBindings.FfiConversation,
            ) -> None:
                conv = self_outer._wrap_conversation(conversation)
                loop.call_soon_threadsafe(queue.put_nowait, conv)

            def on_error(
                self,
                error: NativeBindings.FfiSubscribeError,
            ) -> None:
                loop.call_soon_threadsafe(queue.put_nowait, error)

            def on_close(self) -> None:
                loop.call_soon_threadsafe(stream._end)

        self_outer = self
        callback = Callback()
        stream: AsyncStream[Conversation | NativeBindings.FfiSubscribeError] = AsyncStream(queue)

        async def start() -> None:
            closer = await self._ensure().stream(callback)

            async def close() -> None:
                await closer.end_and_wait()

            stream._closer = close

        asyncio.create_task(start())
        return stream

    def stream_all_messages(
        self,
    ) -> AsyncStream[DecodedMessage[object] | NativeBindings.FfiSubscribeError]:
        """Stream all messages across conversations."""

        queue: asyncio.Queue[object] = asyncio.Queue()
        loop = asyncio.get_event_loop()
        client = self._client

        class Callback(NativeBindings.FfiMessageCallback):
            def on_message(
                self,
                message: NativeBindings.FfiMessage,
            ) -> None:
                decoded = client._decode_message(message)
                loop.call_soon_threadsafe(queue.put_nowait, decoded)

            def on_error(
                self,
                error: NativeBindings.FfiSubscribeError,
            ) -> None:
                loop.call_soon_threadsafe(queue.put_nowait, error)

            def on_close(self) -> None:
                loop.call_soon_threadsafe(stream._end)

        callback = Callback()
        stream: AsyncStream[DecodedMessage[object] | NativeBindings.FfiSubscribeError] = (
            AsyncStream(queue)
        )

        async def start() -> None:
            closer = await self._ensure().stream_all_messages(callback, None)

            async def close() -> None:
                await closer.end_and_wait()

            stream._closer = close

        asyncio.create_task(start())
        return stream

    async def sync(self) -> None:
        """Sync new conversations."""

        await self._ensure().sync()

    async def sync_all_conversations(
        self,
        consent_states: builtins.list[NativeBindings.FfiConsentState] | None = None,
    ) -> None:
        """Sync all conversations."""
        ffi = self._ensure()
        if consent_states is None:
            try:
                signature = inspect.signature(ffi.sync_all_conversations)
            except (TypeError, ValueError):
                await ffi.sync_all_conversations(consent_states)
                return
            if len(signature.parameters) == 0:
                await cast(Any, ffi).sync_all_conversations()
                return
        await ffi.sync_all_conversations(consent_states)

    def _wrap_conversation(self, convo: NativeBindings.FfiConversation) -> Conversation:
        convo_type = convo.conversation_type()
        if convo_type == NativeBindings.FfiConversationType.DM:
            return Dm(self._client, convo)
        return Group(self._client, convo)



"""Utility helpers for XMTP."""

from __future__ import annotations

import re

_HEX_RE = re.compile(r"^(0x)?[0-9a-fA-F]+$")


def is_hex_string(value: str, length: int | None = None) -> bool:
    """Return True if value is a hex string.

    Args:
        value: String to validate.
        length: Optional length (in hex digits, excluding 0x).
    """

    if not _HEX_RE.match(value):
        return False

    normalized = value[2:] if value.startswith("0x") else value
    if length is None:
        return True
    return len(normalized) == length


def hex_to_bytes(value: str) -> bytes:
    """Convert a hex string (with or without 0x prefix) to bytes."""

    normalized = value[2:] if value.startswith("0x") else value
    return bytes.fromhex(normalized)


def coerce_db_encryption_key(value: bytes | str | None) -> bytes | None:
    """Normalize the database encryption key to bytes if provided."""

    if value is None:
        return None
    if isinstance(value, bytes):
        return value
    if is_hex_string(value):
        return hex_to_bytes(value)
    raise ValueError("db_encryption_key must be bytes or hex string")



"""Transaction reference content type for XMTP."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, cast

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _FfiTransactionMetadata(Protocol):
    def __init__(
        self,
        *,
        transaction_type: str,
        currency: str,
        amount: float,
        decimals: int,
        from_address: str,
        to_address: str,
    ) -> None: ...

    transaction_type: str
    currency: str
    amount: float
    decimals: int
    from_address: str
    to_address: str


class _FfiTransactionReference(Protocol):
    def __init__(
        self,
        *,
        namespace: str | None,
        network_id: str,
        reference: str,
        metadata: _FfiTransactionMetadata | None,
    ) -> None: ...

    namespace: str | None
    network_id: str
    reference: str
    metadata: _FfiTransactionMetadata | None


class _Bindings(Protocol):
    FfiTransactionMetadata: type[_FfiTransactionMetadata]
    FfiTransactionReference: type[_FfiTransactionReference]

    def encode_transaction_reference(self, payload: _FfiTransactionReference) -> bytes: ...

    def decode_transaction_reference(self, data: bytes) -> _FfiTransactionReference: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeTransactionReference = ContentTypeId(
    authority_id="xmtp.org",
    type_id="transactionReference",
    version_major=1,
    version_minor=0,
)


@dataclass(slots=True)
class TransactionMetadata:
    """Optional transaction metadata."""

    transaction_type: str
    currency: str
    amount: float
    decimals: int
    from_address: str
    to_address: str


@dataclass(slots=True)
class TransactionReference:
    """Transaction reference content."""

    namespace: str | None
    network_id: str | int
    reference: str
    metadata: TransactionMetadata | None = None


class TransactionReferenceCodec(ContentCodec[TransactionReference]):
    """Codec for transaction references."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeTransactionReference

    def encode(
        self,
        content: TransactionReference,
        registry: CodecRegistry | None = None,
    ) -> EncodedContent:
        metadata = None
        if content.metadata is not None:
            metadata = _bindings().FfiTransactionMetadata(
                transaction_type=content.metadata.transaction_type,
                currency=content.metadata.currency,
                amount=content.metadata.amount,
                decimals=content.metadata.decimals,
                from_address=content.metadata.from_address,
                to_address=content.metadata.to_address,
            )
        payload = _bindings().FfiTransactionReference(
            namespace=content.namespace,
            network_id=str(content.network_id),
            reference=content.reference,
            metadata=metadata,
        )
        encoded = _bindings().encode_transaction_reference(payload)
        return EncodedContent(type_id=self.content_type, parameters={}, content=encoded)

    def decode(
        self,
        content: EncodedContent,
        registry: CodecRegistry | None = None,
    ) -> TransactionReference:
        payload = _bindings().decode_transaction_reference(content.content)
        metadata = None
        if payload.metadata is not None:
            metadata = TransactionMetadata(
                transaction_type=payload.metadata.transaction_type,
                currency=payload.metadata.currency,
                amount=payload.metadata.amount,
                decimals=payload.metadata.decimals,
                from_address=payload.metadata.from_address,
                to_address=payload.metadata.to_address,
            )
        return TransactionReference(
            namespace=payload.namespace,
            network_id=payload.network_id,
            reference=payload.reference,
            metadata=metadata,
        )

    def fallback(self, content: TransactionReference) -> str | None:
        if content.reference:
            return (
                "[Crypto transaction] Use a blockchain explorer to learn more "
                f"using the transaction hash: {content.reference}"
            )
        return "Crypto transaction"

    def should_push(self, content: TransactionReference) -> bool:
        return True


__all__ = [
    "ContentTypeTransactionReference",
    "TransactionMetadata",
    "TransactionReference",
    "TransactionReferenceCodec",
]



"""XMTP Python SDK (unofficial)."""

from xmtp.client import Client
from xmtp.conversation import Conversation, Dm, Group
from xmtp.conversations import Conversations
from xmtp.env import load_client_options_from_env, load_signer_from_env
from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.messages import DecodedMessage
from xmtp.preferences import Preferences
from xmtp.types import ClientOptions, LogLevel, XmtpEnv

__all__ = [
    "Client",
    "ClientOptions",
    "Conversation",
    "Conversations",
    "DecodedMessage",
    "Dm",
    "Group",
    "Identifier",
    "IdentifierKind",
    "LogLevel",
    "load_client_options_from_env",
    "load_signer_from_env",
    "Preferences",
    "XmtpEnv",
]

__version__ = "0.1.5"



"""Name resolution utilities."""

from __future__ import annotations

import asyncio
import json
from collections import OrderedDict
from collections.abc import Awaitable, Callable
from typing import Final, Generic, TypeVar, cast
from urllib.parse import quote
from urllib.request import Request, urlopen

from xmtp.utils import is_hex_string

from xmtp_agent.errors import AgentError


class _MissingType:
    __slots__ = ()


_MISSING: Final = _MissingType()
T = TypeVar("T")


class _LimitedCache(Generic[T]):
    def __init__(self, limit: int = 1000) -> None:
        self._limit = limit
        self._data: OrderedDict[str, T] = OrderedDict()

    def get(self, key: str) -> T | _MissingType:
        if key not in self._data:
            return _MISSING
        self._data.move_to_end(key)
        return self._data[key]

    def set(self, key: str, value: T) -> None:
        self._data[key] = value
        self._data.move_to_end(key)
        if len(self._data) > self._limit:
            self._data.popitem(last=False)


def create_name_resolver(api_key: str | None = None) -> Callable[[str], Awaitable[str | None]]:
    """Create an async name resolver backed by web3.bio."""

    cache: _LimitedCache[str | None] = _LimitedCache(1000)

    async def resolve_name(name: str) -> str | None:
        if is_hex_string(name, length=40):
            return name

        cached = cache.get(name)
        if cached is not _MISSING:
            return cast(str | None, cached)

        def fetch() -> list[dict[str, object]]:
            endpoint = f"https://api.web3.bio/ns/{quote(name)}"
            headers = {"Content-Type": "application/json"}
            if api_key:
                headers["X-API-KEY"] = f"Bearer {api_key}"
            request = Request(endpoint, headers=headers, method="GET")
            with urlopen(request, timeout=10) as response:
                if response.status >= 400:
                    raise AgentError(
                        f'Could not resolve address for name "{name}": '
                        f"{response.status} {response.reason}"
                    )
                data = response.read()
            return cast(list[dict[str, object]], json.loads(data.decode("utf-8")))

        results = await asyncio.to_thread(fetch)
        address_value = results[0].get("address") if results else None
        address = address_value if isinstance(address_value, str) else None
        cache.set(name, address)
        return address

    return resolve_name


__all__ = ["create_name_resolver"]



"""Reaction content type for XMTP."""

from __future__ import annotations

import json
from dataclasses import dataclass
from enum import Enum
from typing import Protocol, cast

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _FfiReactionAction(Protocol):
    ADDED: object
    REMOVED: object


class _FfiReactionSchema(Protocol):
    UNICODE: object
    SHORTCODE: object
    CUSTOM: object


class _FfiReactionPayload(Protocol):
    def __init__(
        self,
        *,
        reference: str,
        reference_inbox_id: str,
        action: object,
        content: str,
        schema: object,
    ) -> None: ...

    reference: str
    reference_inbox_id: str
    action: object
    content: str
    schema: object


class _Bindings(Protocol):
    FfiReactionAction: type[_FfiReactionAction]
    FfiReactionSchema: type[_FfiReactionSchema]
    FfiReactionPayload: type[_FfiReactionPayload]

    def encode_reaction(self, payload: _FfiReactionPayload) -> bytes: ...

    def decode_reaction(self, data: bytes) -> _FfiReactionPayload: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeReaction = ContentTypeId(
    authority_id="xmtp.org",
    type_id="reaction",
    version_major=1,
    version_minor=0,
)


class ReactionAction(str, Enum):
    ADDED = "added"
    REMOVED = "removed"


class ReactionSchema(str, Enum):
    UNICODE = "unicode"
    SHORTCODE = "shortcode"
    CUSTOM = "custom"


@dataclass(slots=True)
class Reaction:
    reference: str
    reference_inbox_id: str | None
    action: ReactionAction
    content: str
    schema: ReactionSchema


class ReactionCodec(ContentCodec[Reaction]):
    """Codec for reaction messages."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeReaction

    def encode(self, content: Reaction, registry: CodecRegistry | None = None) -> EncodedContent:
        ffi_action = (
            _bindings().FfiReactionAction.ADDED
            if content.action == ReactionAction.ADDED
            else _bindings().FfiReactionAction.REMOVED
        )
        ffi_schema = {
            ReactionSchema.UNICODE: _bindings().FfiReactionSchema.UNICODE,
            ReactionSchema.SHORTCODE: _bindings().FfiReactionSchema.SHORTCODE,
            ReactionSchema.CUSTOM: _bindings().FfiReactionSchema.CUSTOM,
        }[content.schema]
        payload = _bindings().FfiReactionPayload(
            reference=content.reference,
            reference_inbox_id=content.reference_inbox_id or "",
            action=ffi_action,
            content=content.content,
            schema=ffi_schema,
        )
        encoded = _bindings().encode_reaction(payload)
        return EncodedContent(type_id=self.content_type, parameters={}, content=encoded)

    def decode(self, content: EncodedContent, registry: CodecRegistry | None = None) -> Reaction:
        decoded_text: str | None = None
        if isinstance(content.content, (bytes, bytearray)):
            try:
                decoded_text = bytes(content.content).decode("utf-8")
            except UnicodeDecodeError:
                decoded_text = None

        if decoded_text:
            try:
                payload = json.loads(decoded_text)
                reference_inbox_id = payload.get(
                    "referenceInboxId", payload.get("reference_inbox_id")
                )
                return Reaction(
                    reference=payload["reference"],
                    reference_inbox_id=reference_inbox_id,
                    action=ReactionAction(payload["action"]),
                    content=payload["content"],
                    schema=ReactionSchema(payload["schema"]),
                )
            except (ValueError, KeyError, TypeError):
                pass

        if decoded_text is not None:
            params = content.parameters
            if "action" in params and "reference" in params and "schema" in params:
                return Reaction(
                    reference=params["reference"],
                    reference_inbox_id=None,
                    action=ReactionAction(params["action"]),
                    content=decoded_text,
                    schema=ReactionSchema(params["schema"]),
                )

        payload = _bindings().decode_reaction(content.content)
        action = (
            ReactionAction.ADDED
            if payload.action == _bindings().FfiReactionAction.ADDED
            else ReactionAction.REMOVED
        )
        schema_map = {
            _bindings().FfiReactionSchema.UNICODE: ReactionSchema.UNICODE,
            _bindings().FfiReactionSchema.SHORTCODE: ReactionSchema.SHORTCODE,
            _bindings().FfiReactionSchema.CUSTOM: ReactionSchema.CUSTOM,
        }
        schema = schema_map[payload.schema]
        reference_inbox_id = payload.reference_inbox_id or None
        return Reaction(
            reference=payload.reference,
            reference_inbox_id=reference_inbox_id,
            action=action,
            content=payload.content,
            schema=schema,
        )

    def fallback(self, content: Reaction) -> str | None:
        if content.action == ReactionAction.ADDED:
            return f'Reacted "{content.content}" to an earlier message'
        if content.action == ReactionAction.REMOVED:
            return f'Removed "{content.content}" from an earlier message'
        return None

    def should_push(self, content: Reaction) -> bool:
        return False


__all__ = ["ContentTypeReaction", "Reaction", "ReactionAction", "ReactionSchema", "ReactionCodec"]



"""Signer helpers for XMTP."""

from xmtp.signers.base import Signer, SignerType
from xmtp.signers.eoa import EoaSigner, create_signer
from xmtp.signers.scw import ScwSigner

__all__ = ["Signer", "SignerType", "EoaSigner", "ScwSigner", "create_signer"]



"""Group updated content type for XMTP."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, TypeAlias, cast

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)

if TYPE_CHECKING:
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    GroupUpdated: TypeAlias = xmtpv3.FfiGroupUpdated
else:
    GroupUpdated: TypeAlias = object


class _Bindings(Protocol):
    def decode_group_updated(self, data: bytes) -> GroupUpdated: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeGroupUpdated = ContentTypeId(
    authority_id="xmtp.org",
    type_id="group_updated",
    version_major=1,
    version_minor=0,
)


class GroupUpdatedCodec(ContentCodec[GroupUpdated]):
    """Codec for group updated messages."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeGroupUpdated

    def encode(
        self, content: GroupUpdated, registry: CodecRegistry | None = None
    ) -> EncodedContent:
        raise NotImplementedError(
            "GroupUpdated messages are system generated and cannot be encoded"
        )

    def decode(
        self, content: EncodedContent, registry: CodecRegistry | None = None
    ) -> GroupUpdated:
        return _bindings().decode_group_updated(content.content)

    def fallback(self, content: GroupUpdated) -> str | None:
        return None

    def should_push(self, content: GroupUpdated) -> bool:
        return False


__all__ = ["ContentTypeGroupUpdated", "GroupUpdated", "GroupUpdatedCodec"]



"""Command router middleware for XMTP agents."""

from __future__ import annotations

from collections.abc import Awaitable, Callable

from xmtp_agent.context import MessageContext
from xmtp_agent.middleware import Middleware

Handler = Callable[[MessageContext], Awaitable[None]]


class CommandRouter:
    """Route slash commands to handlers for text messages."""

    def __init__(self) -> None:
        self._commands: dict[str, Handler] = {}
        self._default: Handler | None = None

    @property
    def command_list(self) -> list[str]:
        return list(self._commands.keys())

    def command(self, command: str, handler: Handler) -> CommandRouter:
        if not command.startswith("/"):
            raise ValueError('Command must start with "/"')
        self._commands[command.lower()] = handler
        return self

    def default(self, handler: Handler) -> CommandRouter:
        self._default = handler
        return self

    async def handle(self, ctx: MessageContext) -> bool:
        if not ctx.is_text():
            return False
        message_text = ctx.message.content
        if not isinstance(message_text, str):
            return False
        parts = message_text.split(" ")
        command = parts[0].lower() if parts else ""
        if not command:
            return False
        if command.startswith("/"):
            handler = self._commands.get(command)
            if handler:
                ctx.message.content = " ".join(parts[1:])
                await handler(ctx)
                return True
        if self._default:
            await self._default(ctx)
            return True
        return False

    def middleware(self) -> Middleware:
        async def _middleware(
            ctx: MessageContext,
            next_handler: Callable[[], Awaitable[None]],
        ) -> None:
            handled = await self.handle(ctx)
            if not handled:
                await next_handler()

        return _middleware


__all__ = ["CommandRouter"]



"""Bindings loader for libxmtp (UniFFI)."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xmtp_bindings import xmtpv3 as NativeBindings

try:
    from xmtp_bindings import xmtpv3 as _xmtpv3
except (ImportError, OSError) as exc:  # pragma: no cover - import guard

    class _MissingBindings:
        def __init__(self, error: Exception) -> None:
            self._error = error

        def __getattr__(self, name: str) -> object:
            raise ImportError(
                "xmtp-bindings is required. Build bindings/python or install the package."
            ) from self._error

    NativeBindings = _MissingBindings(exc)  # type: ignore[assignment]
else:
    NativeBindings = _xmtpv3

__all__ = ["NativeBindings"]



"""Testing utilities for agent development."""

from __future__ import annotations

import asyncio
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Generic, TypeVar

from xmtp.async_stream import AsyncStream
from xmtp.messages import DecodedMessage
from xmtp_content_type_text import ContentTypeText

ContentT = TypeVar("ContentT")


@dataclass(slots=True)
class _MessageOverrides(Generic[ContentT]):
    id: bytes | None = None
    conversation_id: bytes | None = None
    sender_inbox_id: str | None = None
    sent_at: datetime | None = None
    content_type_id: str | None = None
    content: ContentT | None = None


def create_mock_message(content: ContentT, **overrides: object) -> DecodedMessage[ContentT]:
    """Create a decoded message with sensible defaults for tests."""

    opts = _MessageOverrides[ContentT](content=content, **overrides)  # type: ignore[arg-type]
    return DecodedMessage(
        id=opts.id or b"mock-message-id",
        conversation_id=opts.conversation_id or b"test-conversation-id",
        sender_inbox_id=opts.sender_inbox_id or "sender-inbox-id",
        sent_at=opts.sent_at or datetime.now(tz=timezone.utc),
        content=content,
        content_type_id=opts.content_type_id or str(ContentTypeText),
    )


class MockAsyncStream(AsyncStream[ContentT], Generic[ContentT]):
    """Async stream helper with push/end for tests."""

    def __init__(self, values: Iterable[ContentT] | None = None) -> None:
        queue: asyncio.Queue[object] = asyncio.Queue()
        super().__init__(queue)
        if values is not None:
            for value in values:
                queue.put_nowait(value)

    def push(self, value: ContentT) -> None:
        self._queue.put_nowait(value)

    def end(self) -> None:
        self._end()

    async def close(self) -> None:
        self.end()


def serialize_message(message: DecodedMessage[Any]) -> dict[str, Any]:
    """Serialize a decoded message to a JSON-friendly dict."""

    return {
        "id": message.id.hex(),
        "conversation_id": message.conversation_id.hex(),
        "sender_inbox_id": message.sender_inbox_id,
        "sent_at": message.sent_at.isoformat(),
        "content": message.content,
        "content_type_id": message.content_type_id,
    }


def deserialize_message(payload: dict[str, Any]) -> DecodedMessage[Any]:
    """Deserialize a decoded message from a JSON-friendly dict."""

    return DecodedMessage(
        id=bytes.fromhex(payload["id"]),
        conversation_id=bytes.fromhex(payload["conversation_id"]),
        sender_inbox_id=payload["sender_inbox_id"],
        sent_at=datetime.fromisoformat(payload["sent_at"]),
        content=payload["content"],
        content_type_id=payload.get("content_type_id"),
    )


async def record_messages(
    stream: AsyncStream[DecodedMessage[Any]],
    limit: int | None = None,
) -> list[dict[str, Any]]:
    """Record decoded messages from a stream."""

    recorded: list[dict[str, Any]] = []
    async for message in stream:
        recorded.append(serialize_message(message))
        if limit is not None and len(recorded) >= limit:
            break
    return recorded


def replay_messages(records: Iterable[dict[str, Any]]) -> MockAsyncStream[DecodedMessage[Any]]:
    """Replay recorded messages as an async stream."""

    messages = [deserialize_message(record) for record in records]
    stream: MockAsyncStream[DecodedMessage[Any]] = MockAsyncStream(messages)
    stream.end()
    return stream


async def flush_asyncio() -> None:
    """Flush the asyncio event loop for tests."""

    await asyncio.sleep(0)


__all__ = [
    "MockAsyncStream",
    "create_mock_message",
    "deserialize_message",
    "flush_asyncio",
    "record_messages",
    "replay_messages",
    "serialize_message",
]



"""Agent class for XMTP."""

from __future__ import annotations

import asyncio
import inspect
import os
import warnings
from collections import defaultdict
from collections.abc import Awaitable, Callable

from xmtp import Client, ClientOptions, Conversation, DecodedMessage, Dm, Group
from xmtp.async_stream import AsyncStream
from xmtp.bindings import NativeBindings
from xmtp.env import load_client_options_from_env, load_signer_from_env
from xmtp.signers.base import Signer
from xmtp.types import LogLevel

from xmtp_agent.context import ClientContext, ConversationContext, MessageContext
from xmtp_agent.debug import get_installation_info
from xmtp_agent.errors import AgentError, AgentStreamingError
from xmtp_agent.filters import (
    from_self,
    has_content,
    is_group_update,
    is_markdown,
    is_reaction,
    is_read_receipt,
    is_remote_attachment,
    is_reply,
    is_text,
    is_transaction_reference,
    is_wallet_send_calls,
)
from xmtp_agent.middleware import ErrorMiddleware, Middleware

EventPayload = ClientContext | ConversationContext | MessageContext | Exception
EventHandler = Callable[[EventPayload], Awaitable[None] | None]


class _ErrorRegistrar:
    def __init__(self, agent: Agent) -> None:
        self._agent = agent

    def use(self, *error_middlewares: ErrorMiddleware | list[ErrorMiddleware]) -> _ErrorRegistrar:
        for middleware in error_middlewares:
            if isinstance(middleware, list):
                self._agent._error_middlewares.extend(middleware)
            else:
                self._agent._error_middlewares.append(middleware)
        return self


class Agent:
    """Main agent class with an event-driven API."""

    def __init__(self, client: Client) -> None:
        self._client = client
        self._handlers: dict[str, list[EventHandler]] = defaultdict(list)
        self._middlewares: list[Middleware] = []
        self._error_middlewares: list[ErrorMiddleware] = []
        self._errors = _ErrorRegistrar(self)
        self._conversation_stream_handle: (
            AsyncStream[Conversation | NativeBindings.FfiSubscribeError] | None
        ) = None
        self._message_stream_handle: (
            AsyncStream[DecodedMessage[object] | NativeBindings.FfiSubscribeError] | None
        ) = None
        self._conversation_stream: asyncio.Task[None] | None = None
        self._message_stream: asyncio.Task[None] | None = None
        self._running = False

    @classmethod
    async def create(
        cls,
        signer: Signer,
        options: ClientOptions | None = None,
    ) -> Agent:
        """Create an agent with a signer."""

        opts = options or ClientOptions()
        if opts.app_version is None:
            opts.app_version = "agent-sdk/alpha"
        if not opts.disable_device_sync:
            opts.disable_device_sync = True

        if os.getenv("XMTP_FORCE_DEBUG"):
            opts.debug_events_enabled = True
            opts.structured_logging = True
            level = os.getenv("XMTP_FORCE_DEBUG_LEVEL")
            try:
                opts.logging_level = LogLevel(level) if level else LogLevel.WARN
            except ValueError:
                opts.logging_level = LogLevel.WARN

        client = await Client.create(signer, opts)
        info = await get_installation_info(client)
        if info.total_installations > 1 and info.is_most_recent:
            warnings.warn(
                "You have multiple installations. Installation ID "
                f'"{info.installation_id}" is the most recent. Persist and reload your '
                "installation data to avoid losing access. If you exceed the installation "
                "limit, your agent will stop working. See "
                "https://docs.xmtp.org/agents/build-agents/local-database"
                "#installation-limits-and-revocation-rules",
                UserWarning,
                stacklevel=2,
            )
        return cls(client)

    @classmethod
    async def create_from_env(cls, options: ClientOptions | None = None) -> Agent:
        """Create an agent from environment variables."""

        signer = load_signer_from_env()
        opts = load_client_options_from_env(options)
        return await cls.create(signer, opts)

    def on(
        self,
        event: str,
        handler: EventHandler | None = None,
    ) -> EventHandler | Callable[[EventHandler], EventHandler]:
        """Register an event handler."""

        if handler is None:

            def decorator(fn: EventHandler) -> EventHandler:
                self._handlers[event].append(fn)
                return fn

            return decorator

        self._handlers[event].append(handler)
        return handler

    def use(self, *middlewares: Middleware | list[Middleware]) -> Agent:
        """Register middleware for message handling."""

        for middleware in middlewares:
            if isinstance(middleware, list):
                self._middlewares.extend(middleware)
            else:
                self._middlewares.append(middleware)
        return self

    @property
    def errors(self) -> _ErrorRegistrar:
        """Return an object to register error middleware."""

        return self._errors

    @property
    def client(self) -> Client:
        return self._client

    async def start(self) -> None:
        """Start the agent."""

        if self._running:
            return
        self._running = True

        await self._emit("start", ClientContext(self._client))

        self._conversation_stream = asyncio.create_task(self._consume_conversations())
        self._message_stream = asyncio.create_task(self._consume_messages())

    async def stop(self) -> None:
        """Stop the agent."""

        self._running = False
        await self._stop_streams()
        await self._emit("stop", ClientContext(self._client))

    async def _stop_streams(self) -> None:
        if self._conversation_stream_handle is not None:
            await self._conversation_stream_handle.close()
            self._conversation_stream_handle = None
        if self._message_stream_handle is not None:
            await self._message_stream_handle.close()
            self._message_stream_handle = None
        tasks = [task for task in (self._conversation_stream, self._message_stream) if task]
        for task in tasks:
            task.cancel()
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)
        self._conversation_stream = None
        self._message_stream = None

    async def _consume_conversations(self) -> None:
        try:
            stream: AsyncStream[Conversation | NativeBindings.FfiSubscribeError] = (
                self._client.conversations.stream()
            )
            self._conversation_stream_handle = stream
            async for item in stream:
                if not self._running:
                    break
                if isinstance(item, NativeBindings.FfiSubscribeError):
                    raise AgentStreamingError(str(item))  # pragma: no cover
                await self._handle_conversation(item)
        except asyncio.CancelledError:
            return
        except Exception as error:
            await self._handle_stream_error(error)
        finally:
            if self._conversation_stream_handle is not None:
                await self._conversation_stream_handle.close()
                self._conversation_stream_handle = None

    async def _consume_messages(self) -> None:
        try:
            stream: AsyncStream[DecodedMessage[object] | NativeBindings.FfiSubscribeError] = (
                self._client.conversations.stream_all_messages()
            )
            self._message_stream_handle = stream
            async for item in stream:
                if not self._running:
                    break
                if isinstance(item, NativeBindings.FfiSubscribeError):
                    raise AgentStreamingError(str(item))  # pragma: no cover
                await self._handle_message(item)
        except asyncio.CancelledError:
            return
        except Exception as error:
            await self._handle_stream_error(error)
        finally:
            if self._message_stream_handle is not None:
                await self._message_stream_handle.close()
                self._message_stream_handle = None

    async def _handle_stream_error(self, error: Exception) -> None:
        await self._stop_streams()
        recovered = await self._run_error_chain(error, ClientContext(self._client))
        if recovered:
            self._running = False
            await self.start()

    async def _handle_conversation(self, conversation: Conversation) -> None:
        context = ConversationContext(conversation, self._client)
        await self._emit("conversation", context)
        if isinstance(conversation, Dm):
            await self._emit("dm", context)
        elif isinstance(conversation, Group):
            await self._emit("group", context)

    async def _handle_message(self, message: DecodedMessage[object]) -> None:
        if not has_content(message):
            return
        if from_self(message, self._client):
            return

        conversation = await self._client.conversations.get_conversation_by_id(
            message.conversation_id
        )
        if conversation is None:
            raise AgentError(
                f"Failed to process message {message.id.hex()} for conversation "
                f"{message.conversation_id.hex()}: conversation not found."
            )

        context = MessageContext(message, conversation, self._client)
        topic = self._topic_for_message(message)
        await self._run_middleware_chain(context, topic)

    def _topic_for_message(self, message: DecodedMessage[object]) -> str:
        if is_text(message):
            return "text"
        if is_markdown(message):
            return "markdown"
        if is_reaction(message):
            return "reaction"
        if is_reply(message):
            return "reply"
        if is_remote_attachment(message):
            return "attachment"
        if is_read_receipt(message):
            return "read-receipt"
        if is_group_update(message):
            return "group-update"
        if is_transaction_reference(message):
            return "transaction-reference"
        if is_wallet_send_calls(message):
            return "wallet-send-calls"
        return "unknown_message"

    async def _run_middleware_chain(self, context: MessageContext, topic: str) -> None:
        async def final_emit() -> None:
            try:
                await self._emit(topic, context)
                await self._emit("message", context)
            except Exception as error:
                await self._run_error_chain(error, context)

        def wrap(
            middleware: Middleware,
            next_handler: Callable[[], Awaitable[None]],
        ) -> Callable[[], Awaitable[None]]:
            async def _wrapped() -> None:
                try:
                    await middleware(context, next_handler)
                except Exception as error:
                    resume = await self._run_error_chain(error, context)
                    if resume:
                        await next_handler()

            return _wrapped

        next_handler: Callable[[], Awaitable[None]] = final_emit
        for middleware in reversed(self._middlewares):
            next_handler = wrap(middleware, next_handler)

        await next_handler()

    async def _run_error_handler(
        self,
        handler: ErrorMiddleware,
        context: MessageContext | ClientContext,
        error: Exception,
    ) -> tuple[str, Exception | None]:
        settled = False
        outcome = "stopped"
        next_error: Exception | None = None

        async def next_handler(next_exc: Exception | None = None) -> None:
            nonlocal settled, outcome, next_error
            if settled:
                return
            settled = True
            if next_exc is None:
                outcome = "handled"
            else:
                outcome = "continue"
                next_error = next_exc

        try:
            result = handler(error, context, next_handler)
            if inspect.isawaitable(result):
                await result
        except Exception as exc:  # pragma: no cover - defensive
            if not settled:
                return "continue", exc
        return outcome, next_error

    async def _run_error_chain(
        self,
        error: Exception,
        context: MessageContext | ClientContext,
    ) -> bool:
        chain = [*self._error_middlewares, self._default_error_handler]
        current_error: Exception = error

        for handler in chain:
            outcome, next_error = await self._run_error_handler(handler, context, current_error)
            if outcome == "handled":
                return True
            if outcome == "stopped":
                return False
            if outcome == "continue" and next_error is not None:  # pragma: no branch
                current_error = next_error
        return False

    async def _default_error_handler(
        self,
        error: Exception,
        context: MessageContext | ClientContext,
        next_handler: Callable[[Exception | None], Awaitable[None]],
    ) -> None:
        await self._emit("unhandled_error", error)

    async def _emit(self, event: str, arg: EventPayload) -> None:
        handlers = self._handlers.get(event, [])
        for handler in handlers:
            result = handler(arg)
            if inspect.isawaitable(result):
                await result



"""Text content type for XMTP."""

from __future__ import annotations

from enum import Enum
from typing import Protocol

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _Bindings(Protocol):
    def encode_text(self, text: str) -> bytes: ...

    def decode_text(self, data: bytes) -> str: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return xmtpv3  # pragma: no cover - requires native bindings


ContentTypeText = ContentTypeId(
    authority_id="xmtp.org",
    type_id="text",
    version_major=1,
    version_minor=0,
)


class Encoding(str, Enum):
    UTF8 = "UTF-8"
    UNKNOWN = "unknown"


class TextCodec(ContentCodec[str]):
    """Codec for plain text messages."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeText

    def encode(self, content: str, registry: CodecRegistry | None = None) -> EncodedContent:
        encoded = _bindings().encode_text(content)
        return EncodedContent(
            type_id=self.content_type,
            parameters={"encoding": Encoding.UTF8.value},
            content=encoded,
        )

    def decode(self, content: EncodedContent, registry: CodecRegistry | None = None) -> str:
        encoding = content.parameters.get("encoding")
        if encoding is None:
            raise ValueError("Missing encoding for text content")
        if encoding.upper() != Encoding.UTF8.value:
            raise ValueError(f"unrecognized encoding {encoding}")
        if not isinstance(content.content, (bytes, bytearray)):
            raise TypeError("Text content payload must be bytes")
        return _bindings().decode_text(bytes(content.content))

    def fallback(self, content: str) -> str | None:
        return None

    def should_push(self, content: str) -> bool:
        return True


__all__ = ["ContentTypeText", "Encoding", "TextCodec"]



"""Async stream helper."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Awaitable, Callable
from typing import Generic, TypeVar

T = TypeVar("T")

_END = object()


class AsyncStream(AsyncIterator[T], Generic[T]):
    """Async stream backed by an asyncio queue."""

    def __init__(
        self,
        queue: asyncio.Queue[object],
        closer: Callable[[], Awaitable[None]] | None = None,
    ) -> None:
        self._queue = queue
        self._closer = closer

    def __aiter__(self) -> AsyncStream[T]:
        return self

    async def __anext__(self) -> T:
        item = await self._queue.get()
        if item is _END:
            raise StopAsyncIteration
        return item  # type: ignore[return-value]

    async def close(self) -> None:
        """Close the underlying stream if possible."""

        if self._closer is None:
            return
        await self._closer()

    def _end(self) -> None:
        """Signal end-of-stream to consumers."""

        self._queue.put_nowait(_END)



"""Read receipt content type for XMTP."""

from __future__ import annotations

from typing import Protocol, TypeAlias, cast

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _FfiReadReceipt(Protocol):
    def __init__(self) -> None: ...


class _Bindings(Protocol):
    FfiReadReceipt: type[_FfiReadReceipt]

    def encode_read_receipt(self, payload: _FfiReadReceipt) -> bytes: ...

    def decode_read_receipt(self, data: bytes) -> object: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeReadReceipt = ContentTypeId(
    authority_id="xmtp.org",
    type_id="readReceipt",
    version_major=1,
    version_minor=0,
)


ReadReceipt: TypeAlias = dict[str, object]


class ReadReceiptCodec(ContentCodec[ReadReceipt]):
    """Codec for read receipt messages."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeReadReceipt

    def encode(self, content: ReadReceipt, registry: CodecRegistry | None = None) -> EncodedContent:
        encoded = _bindings().encode_read_receipt(_bindings().FfiReadReceipt())
        return EncodedContent(type_id=self.content_type, parameters={}, content=encoded)

    def decode(self, content: EncodedContent, registry: CodecRegistry | None = None) -> ReadReceipt:
        _bindings().decode_read_receipt(content.content)
        return {}

    def fallback(self, content: ReadReceipt) -> str | None:
        return None

    def should_push(self, content: ReadReceipt) -> bool:
        return False


__all__ = ["ContentTypeReadReceipt", "ReadReceipt", "ReadReceiptCodec"]



"""Recipient resolution helpers for agents."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass

from xmtp import Client
from xmtp.bindings import NativeBindings
from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.utils import is_hex_string


@dataclass(slots=True)
class ResolvedRecipient:
    """Resolved recipient details."""

    address: str | None
    inbox_id: str | None
    identifier: Identifier | None


def _normalize_hex(value: str) -> str:
    normalized = value[2:] if value.lower().startswith("0x") else value
    return normalized.lower()


def _normalize_address(value: str) -> str:
    if not is_hex_string(value, length=40):
        raise ValueError(f'Invalid address: "{value}"')
    normalized = value if value.lower().startswith("0x") else f"0x{value}"
    return normalized


def _is_inbox_id(value: str) -> bool:
    return is_hex_string(value, length=64)


def _address_from_inbox_state(state: object) -> str | None:
    identities = []
    recovery = getattr(state, "recovery_identity", None)
    if recovery is not None:
        identities.append(recovery)
    identities.extend(getattr(state, "account_identities", []) or [])

    for identity in identities:
        candidate = getattr(identity, "identifier", None)
        if isinstance(candidate, str) and is_hex_string(candidate, length=40):
            return _normalize_address(candidate)
    return None


async def resolve_recipient(
    client: Client,
    recipient: str | Identifier,
    *,
    name_resolver: Callable[[str], Awaitable[str | None]] | None = None,
    consent_state: NativeBindings.FfiConsentState | None = None,
) -> ResolvedRecipient:
    """Resolve an inbox ID or address for a recipient.

    Args:
        client: XMTP client to use for inbox resolution and consent.
        recipient: Inbox ID, address, ENS name, or Identifier.
        name_resolver: Optional async resolver for non-hex names (e.g. ENS).
        consent_state: Optional consent state to apply to the resolved inbox ID.
    """

    address: str | None = None
    inbox_id: str | None = None
    identifier: Identifier | None = None

    if isinstance(recipient, Identifier):
        identifier = recipient
        if identifier.kind == IdentifierKind.ETHEREUM:
            address = _normalize_address(identifier.value)
            identifier = Identifier(kind=IdentifierKind.ETHEREUM, value=address)
        inbox_id = await client.get_inbox_id_by_identifier(identifier)
        if inbox_id is None:
            raise ValueError(f'No inbox id found for identifier "{identifier.value}"')
    else:
        target = recipient.strip()
        if _is_inbox_id(target):
            inbox_id = _normalize_hex(target)
            try:
                state = await client.preferences.get_latest_inbox_state(inbox_id)
            except Exception:
                state = None
            if state is not None:
                address = _address_from_inbox_state(state)
                if address:
                    identifier = Identifier(kind=IdentifierKind.ETHEREUM, value=address)
        else:
            if is_hex_string(target, length=40):
                address = _normalize_address(target)
            else:
                if name_resolver is None:
                    raise ValueError("Name resolver required for non-hex recipient")
                resolved = await name_resolver(target)
                if resolved is None:
                    raise ValueError(f'Could not resolve address for "{target}"')
                address = _normalize_address(resolved)
            identifier = Identifier(kind=IdentifierKind.ETHEREUM, value=address)
            inbox_id = await client.get_inbox_id_by_identifier(identifier)
            if inbox_id is None:
                raise ValueError(f'No inbox id found for address "{address}"')

    if consent_state is not None:
        record = NativeBindings.FfiConsent(
            entity_type=NativeBindings.FfiConsentEntityType.INBOX_ID,
            state=consent_state,
            entity=inbox_id,
        )
        await client.preferences.set_consent_states([record])

    return ResolvedRecipient(address=address, inbox_id=inbox_id, identifier=identifier)


__all__ = ["ResolvedRecipient", "resolve_recipient"]



"""Message filter helpers."""

from __future__ import annotations

from typing import Any, TypeVar

from xmtp import Client, Conversation, DecodedMessage, Dm, Group
from xmtp_content_type_group_updated import ContentTypeGroupUpdated
from xmtp_content_type_markdown import ContentTypeMarkdown
from xmtp_content_type_primitives import ContentCodec
from xmtp_content_type_reaction import ContentTypeReaction
from xmtp_content_type_read_receipt import ContentTypeReadReceipt
from xmtp_content_type_remote_attachment import ContentTypeRemoteAttachment
from xmtp_content_type_reply import ContentTypeReply
from xmtp_content_type_text import ContentTypeText
from xmtp_content_type_transaction_reference import ContentTypeTransactionReference
from xmtp_content_type_wallet_send_calls import ContentTypeWalletSendCalls

ContentT = TypeVar("ContentT")


def from_self(message: DecodedMessage[Any], client: Client) -> bool:
    return message.sender_inbox_id == client.inbox_id


def has_content(message: DecodedMessage[Any]) -> bool:
    return message.content is not None


def is_dm(conversation: Conversation) -> bool:
    return isinstance(conversation, Dm)


def is_group(conversation: Conversation) -> bool:
    return isinstance(conversation, Group)


def is_group_update(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeGroupUpdated)


def is_markdown(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeMarkdown)


def is_reaction(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeReaction)


def is_read_receipt(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeReadReceipt)


def is_remote_attachment(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeRemoteAttachment)


def is_reply(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeReply)


def is_text(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeText)


def is_transaction_reference(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeTransactionReference)


def is_wallet_send_calls(message: DecodedMessage[Any]) -> bool:
    return message.content_type_id == str(ContentTypeWalletSendCalls)


def uses_codec(message: DecodedMessage[Any], codec_class: type[ContentCodec[Any]]) -> bool:
    return message.content_type_id == str(codec_class().content_type)


class _Filter:
    from_self = staticmethod(from_self)
    has_content = staticmethod(has_content)
    is_dm = staticmethod(is_dm)
    is_group = staticmethod(is_group)
    is_group_update = staticmethod(is_group_update)
    is_markdown = staticmethod(is_markdown)
    is_reaction = staticmethod(is_reaction)
    is_read_receipt = staticmethod(is_read_receipt)
    is_remote_attachment = staticmethod(is_remote_attachment)
    is_reply = staticmethod(is_reply)
    is_text = staticmethod(is_text)
    is_transaction_reference = staticmethod(is_transaction_reference)
    is_wallet_send_calls = staticmethod(is_wallet_send_calls)
    uses_codec = staticmethod(uses_codec)


filter = _Filter()
f = filter

__all__ = [
    "filter",
    "f",
    "from_self",
    "has_content",
    "is_dm",
    "is_group",
    "is_group_update",
    "is_markdown",
    "is_reaction",
    "is_read_receipt",
    "is_remote_attachment",
    "is_reply",
    "is_text",
    "is_transaction_reference",
    "is_wallet_send_calls",
    "uses_codec",
]



"""Context objects for agent handlers."""

from __future__ import annotations

from xmtp import Client, Conversation, DecodedMessage, Dm, Group
from xmtp.bindings import NativeBindings
from xmtp_content_type_markdown import ContentTypeMarkdown
from xmtp_content_type_primitives import ContentCodec, ContentTypeId
from xmtp_content_type_reaction import (
    ContentTypeReaction,
    Reaction,
    ReactionAction,
    ReactionSchema,
)
from xmtp_content_type_remote_attachment import ContentTypeRemoteAttachment, RemoteAttachment
from xmtp_content_type_reply import ContentTypeReply, Reply
from xmtp_content_type_text import ContentTypeText


class ClientContext:
    """Context passed to lifecycle handlers."""

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    def get_client_address(self) -> str | None:
        identifier = self._client.account_identifier
        return identifier.value if identifier else None


class ConversationContext(ClientContext):
    """Context passed to conversation handlers."""

    def __init__(self, conversation: Conversation, client: Client) -> None:
        super().__init__(client)
        self._conversation = conversation

    def is_dm(self) -> bool:
        return isinstance(self._conversation, Dm)

    def is_group(self) -> bool:
        return isinstance(self._conversation, Group)

    async def send_markdown(self, markdown: str) -> None:
        await self._conversation.send(markdown, ContentTypeMarkdown)

    async def send_text(self, text: str) -> None:
        await self._conversation.send(text)

    async def send_remote_attachment(self, attachment: RemoteAttachment) -> None:
        await self._conversation.send(attachment, ContentTypeRemoteAttachment)

    @property
    def conversation(self) -> Conversation:
        return self._conversation

    @property
    def is_allowed(self) -> bool:
        return self._conversation.consent_state == NativeBindings.FfiConsentState.ALLOWED

    @property
    def is_denied(self) -> bool:
        return self._conversation.consent_state == NativeBindings.FfiConsentState.DENIED

    @property
    def is_unknown(self) -> bool:
        return self._conversation.consent_state == NativeBindings.FfiConsentState.UNKNOWN


class MessageContext(ConversationContext):
    """Context passed to message handlers."""

    def __init__(
        self,
        message: DecodedMessage[object],
        conversation: Conversation,
        client: Client,
    ) -> None:
        super().__init__(conversation=conversation, client=client)
        self._message = message

    def uses_codec(self, codec_class: type[ContentCodec[object]]) -> bool:
        return self._message.content_type_id == str(codec_class().content_type)

    def is_markdown(self) -> bool:
        return self._message.content_type_id == str(ContentTypeMarkdown)

    def is_text(self) -> bool:
        return self._message.content_type_id == str(ContentTypeText)

    def is_reply(self) -> bool:
        return self._message.content_type_id == str(ContentTypeReply)

    def is_reaction(self) -> bool:
        return self._message.content_type_id == str(ContentTypeReaction)

    def is_remote_attachment(self) -> bool:
        return self._message.content_type_id == str(ContentTypeRemoteAttachment)

    async def send_reaction(
        self,
        content: str,
        schema: ReactionSchema = ReactionSchema.UNICODE,
    ) -> None:
        reaction = Reaction(
            action=ReactionAction.ADDED,
            reference=self._message.id.hex(),
            reference_inbox_id=self._message.sender_inbox_id,
            schema=schema,
            content=content,
        )
        await self._conversation.send(reaction, ContentTypeReaction)

    async def _send_reply(self, text: str, content_type: ContentTypeId) -> None:
        reply = Reply(
            reference=self._message.id.hex(),
            reference_inbox_id=self._message.sender_inbox_id,
            content_type=content_type,
            content=text,
        )
        await self._conversation.send(reply, ContentTypeReply)

    async def send_text_reply(self, text: str) -> None:
        await self._send_reply(text, ContentTypeText)

    async def send_markdown_reply(self, markdown: str) -> None:
        await self._send_reply(markdown, ContentTypeMarkdown)

    async def get_sender_address(self) -> str | None:
        states = await self.client.preferences.inbox_state_from_inbox_ids(
            [self._message.sender_inbox_id],
            refresh_from_network=False,
        )
        if not states:
            return None
        identifiers = states[0].account_identities
        if not identifiers:
            return None
        return identifiers[0].identifier

    @property
    def message(self) -> DecodedMessage[object]:
        return self._message



"""Agent SDK errors."""

from __future__ import annotations


class AgentError(Exception):
    """Base error for the agent SDK."""


class AgentStreamingError(AgentError):
    """Error raised when streaming fails."""



"""Error types for XMTP."""

from __future__ import annotations


class XmtpError(Exception):
    """Base class for XMTP errors."""


class CodecNotFoundError(XmtpError):
    """Raised when a codec is missing for a content type."""

    def __init__(self, content_type: str) -> None:
        super().__init__(f'Codec not found for "{content_type}" content type')


class MissingContentTypeError(XmtpError):
    """Raised when content type is required but missing."""

    def __init__(self) -> None:
        super().__init__("Content type is required when sending non-text content")


class SignerUnavailableError(XmtpError):
    """Raised when a signer is required but not provided."""

    def __init__(self) -> None:
        super().__init__("Signer unavailable, use Client.create with a signer")


class ClientNotInitializedError(XmtpError):
    """Raised when client is used before initialization."""

    def __init__(self) -> None:
        super().__init__("Client not initialized, use Client.create or Client.build")


class AccountAlreadyAssociatedError(XmtpError):
    """Raised when account is already associated with an inbox."""

    def __init__(self, inbox_id: str) -> None:
        super().__init__(f"Account already associated with inbox {inbox_id}")


class InboxReassignError(XmtpError):
    """Raised when inbox reassignment is disallowed."""

    def __init__(self) -> None:
        super().__init__(
            "Unable to create add account signature text, allow_inbox_reassign must be true"
        )


class StreamFailedError(XmtpError):
    """Raised when a stream fails after retries."""

    def __init__(self, retry_attempts: int) -> None:
        times = "times" if retry_attempts != 1 else "time"
        super().__init__(f"Stream failed, retried {retry_attempts} {times}")


class StreamInvalidRetryAttemptsError(XmtpError):
    """Raised when retry attempts are invalid."""

    def __init__(self) -> None:
        super().__init__("Stream retry attempts must be greater than 0")


class NotImplementedXmtpError(XmtpError):
    """Raised when a feature has not been implemented yet."""


class InvalidGroupMembershipChangeError(XmtpError):
    """Raised when group membership changes are invalid."""

    def __init__(self, message_id: str) -> None:
        super().__init__(f"Invalid group membership change for message {message_id}")


class DatabaseOpenError(XmtpError):
    """Raised when the local XMTP database cannot be opened."""

    def __init__(self, db_path: str | None, detail: str | None = None) -> None:
        message = "Failed to open the local XMTP database."
        if db_path:
            message = f'{message} Path: "{db_path}".'
        message = (
            f"{message} If this is a SQLCipher error, verify that "
            "XMTP_DB_ENCRYPTION_KEY is stable and hex-encoded. "
            "To recover from a bad key or corrupted DB, delete the .db3 file "
            "or point XMTP_DB_DIRECTORY/ClientOptions.db_path to a fresh location."
        )
        if detail:
            message = f"{message} Root error: {detail}"
        super().__init__(message)



"""Smart contract wallet signer."""

from __future__ import annotations

from collections.abc import Awaitable, Callable

from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.signers.base import Signer, SignerType


class ScwSigner(Signer):
    """Signer for smart contract wallets."""

    def __init__(
        self,
        address: str,
        sign_message: Callable[[bytes], Awaitable[bytes]],
        chain_id: int,
        block_number: int | None = None,
    ) -> None:
        self._address = address
        self._sign_message = sign_message
        self._chain_id = chain_id
        self._block_number = block_number
        self.type = SignerType.SCW

    async def get_address(self) -> str:
        """Return the wallet address for this signer."""

        return self._address

    async def get_identifier(self) -> Identifier:
        """Return the XMTP identifier for this signer."""

        return Identifier(kind=IdentifierKind.ETHEREUM, value=self._address)

    async def sign_message(self, message: bytes) -> bytes:
        """Sign a message and return the signature bytes."""

        return await self._sign_message(message)

    async def get_chain_id(self) -> int:
        """Return the chain ID for SCW signatures."""

        return self._chain_id

    async def get_block_number(self) -> int | None:
        """Return the block number for SCW signatures, if any."""

        return self._block_number



"""Wallet send calls content type for XMTP."""

from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from typing import Protocol, cast

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _FfiWalletCallMetadata(Protocol):
    def __init__(
        self,
        *,
        description: str,
        transaction_type: str,
        extra: dict[str, str],
    ) -> None: ...

    description: str
    transaction_type: str
    extra: dict[str, str]


class _FfiWalletCall(Protocol):
    def __init__(
        self,
        *,
        to: str | None,
        data: str | None,
        value: str | None,
        gas: str | None,
        metadata: _FfiWalletCallMetadata | None,
    ) -> None: ...

    to: str | None
    data: str | None
    value: str | None
    gas: str | None
    metadata: _FfiWalletCallMetadata | None


class _FfiWalletSendCalls(Protocol):
    def __init__(
        self,
        *,
        version: str,
        chain_id: str,
        _from: str,
        calls: list[_FfiWalletCall],
        capabilities: dict[str, str] | None,
    ) -> None: ...

    version: str
    chain_id: str
    _from: str
    calls: list[_FfiWalletCall]
    capabilities: dict[str, str] | None


class _Bindings(Protocol):
    FfiWalletCallMetadata: type[_FfiWalletCallMetadata]
    FfiWalletCall: type[_FfiWalletCall]
    FfiWalletSendCalls: type[_FfiWalletSendCalls]

    def encode_wallet_send_calls(self, payload: _FfiWalletSendCalls) -> bytes: ...

    def decode_wallet_send_calls(self, data: bytes) -> _FfiWalletSendCalls: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeWalletSendCalls = ContentTypeId(
    authority_id="xmtp.org",
    type_id="walletSendCalls",
    version_major=1,
    version_minor=0,
)


@dataclass(slots=True)
class WalletCallMetadata:
    """Metadata describing a wallet call."""

    description: str
    transaction_type: str
    extra: dict[str, str]


@dataclass(slots=True)
class WalletCall:
    """Wallet call fields."""

    to: str | None
    data: str | None
    value: str | None
    gas: str | None
    metadata: WalletCallMetadata | None = None


@dataclass(slots=True)
class WalletSendCalls:
    """Wallet send calls payload."""

    version: str
    chain_id: str
    from_address: str
    calls: list[WalletCall]
    capabilities: dict[str, str] | None = None


class WalletSendCallsCodec(ContentCodec[WalletSendCalls]):
    """Codec for wallet send calls."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeWalletSendCalls

    def encode(
        self,
        content: WalletSendCalls,
        registry: CodecRegistry | None = None,
    ) -> EncodedContent:
        calls = []
        for call in content.calls:
            metadata = None
            if call.metadata is not None:
                metadata = _bindings().FfiWalletCallMetadata(
                    description=call.metadata.description,
                    transaction_type=call.metadata.transaction_type,
                    extra=call.metadata.extra,
                )
            calls.append(
                _bindings().FfiWalletCall(
                    to=call.to,
                    data=call.data,
                    value=call.value,
                    gas=call.gas,
                    metadata=metadata,
                )
            )
        payload = _bindings().FfiWalletSendCalls(
            version=content.version,
            chain_id=content.chain_id,
            _from=content.from_address,
            calls=calls,
            capabilities=content.capabilities,
        )
        encoded = _bindings().encode_wallet_send_calls(payload)
        return EncodedContent(type_id=self.content_type, parameters={}, content=encoded)

    def decode(
        self,
        content: EncodedContent,
        registry: CodecRegistry | None = None,
    ) -> WalletSendCalls:
        payload = _bindings().decode_wallet_send_calls(content.content)
        calls: list[WalletCall] = []
        for call in payload.calls:
            metadata = None
            if call.metadata is not None:
                metadata = WalletCallMetadata(
                    description=call.metadata.description,
                    transaction_type=call.metadata.transaction_type,
                    extra=call.metadata.extra,
                )
            calls.append(
                WalletCall(
                    to=call.to,
                    data=call.data,
                    value=call.value,
                    gas=call.gas,
                    metadata=metadata,
                )
            )
        return WalletSendCalls(
            version=payload.version,
            chain_id=payload.chain_id,
            from_address=payload._from,
            calls=calls,
            capabilities=payload.capabilities,
        )

    def fallback(self, content: WalletSendCalls) -> str | None:
        return f"[Transaction request generated]: {json.dumps(asdict(content))}"

    def should_push(self, content: WalletSendCalls) -> bool:
        return True


__all__ = [
    "ContentTypeWalletSendCalls",
    "WalletCall",
    "WalletCallMetadata",
    "WalletSendCalls",
    "WalletSendCallsCodec",
]



"""Attachment utilities for agents."""

from __future__ import annotations

import asyncio
import hashlib
from typing import cast
from urllib.request import Request, urlopen

from xmtp_content_type_remote_attachment import RemoteAttachment


def _sha256_hex(data: bytes) -> str:
    return hashlib.sha256(data).hexdigest()


async def download_remote_attachment(remote_attachment: RemoteAttachment) -> bytes:
    """Download a remote attachment and verify its digest."""

    def fetch() -> bytes:
        request = Request(remote_attachment.url, method="GET")
        with urlopen(request, timeout=30) as response:
            if response.status >= 400:
                raise ValueError(
                    f"Unable to fetch remote attachment: {response.status} {response.reason}"
                )
            return cast(bytes, response.read())

    payload = cast(bytes, await asyncio.to_thread(fetch))
    digest = _sha256_hex(payload)
    if digest.lower() != remote_attachment.content_digest.lower():
        raise ValueError("Remote attachment digest does not match payload")
    return payload


__all__ = ["download_remote_attachment"]



"""Conversation models."""

from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast

from xmtp_content_type_primitives import ContentTypeId

from xmtp.bindings import NativeBindings
from xmtp.errors import MissingContentTypeError
from xmtp.identifiers import Identifier, IdentifierKind

if TYPE_CHECKING:
    from xmtp.client import Client


@dataclass(slots=True)
class _SendMessageOpts:
    should_push: bool


def _default_send_opts(
    should_push: bool = True,
) -> NativeBindings.FfiSendMessageOpts | _SendMessageOpts:
    try:
        return NativeBindings.FfiSendMessageOpts(should_push=should_push)
    except Exception:
        return _SendMessageOpts(should_push=should_push)


def _identifier_to_ffi(identifier: Identifier) -> NativeBindings.FfiIdentifier:
    kind = {
        IdentifierKind.ETHEREUM: NativeBindings.FfiIdentifierKind.ETHEREUM,
        IdentifierKind.PASSKEY: NativeBindings.FfiIdentifierKind.PASSKEY,
    }[identifier.kind]
    return NativeBindings.FfiIdentifier(identifier=identifier.value, identifier_kind=kind)


class Conversation:
    """Base conversation class."""

    def __init__(self, client: Client, ffi_conversation: NativeBindings.FfiConversation) -> None:
        self._client = client
        self._ffi = ffi_conversation

    @property
    def id(self) -> bytes:
        """Conversation identifier bytes."""

        return self._ffi.id()

    @property
    def consent_state(self) -> NativeBindings.FfiConsentState | None:
        """Return the conversation consent state if available."""

        try:
            return self._ffi.consent_state()
        except Exception:
            return None

    async def update_consent_state(self, state: NativeBindings.FfiConsentState) -> None:
        """Update the conversation consent state."""

        result = cast(Any, self._ffi).update_consent_state(state)
        if inspect.isawaitable(result):
            await result

    async def send(
        self,
        content: object,
        content_type: ContentTypeId | str | None = None,
    ) -> bytes:
        """Send a message in the conversation."""

        if isinstance(content, str) and content_type is None:
            return await self._ffi.send_text(content)

        if content_type is None:
            raise MissingContentTypeError()

        encoded, opts = self._client.prepare_for_send(content, content_type)
        return await self._ffi.send(encoded, cast("NativeBindings.FfiSendMessageOpts", opts))


class Dm(Conversation):
    """Direct message conversation."""

    @property
    def peer_inbox_id(self) -> str | None:
        """Inbox ID of the DM peer, if available."""

        return self._ffi.dm_peer_inbox_id()


class Group(Conversation):
    """Group conversation."""

    async def add_members(self, inbox_ids: list[str]) -> None:
        """Add members to the group by inbox ID."""

        await self._ffi.add_members_by_inbox_id(inbox_ids)

    async def remove_members(self, inbox_ids: list[str]) -> None:
        """Remove members from the group by inbox ID."""

        await self._ffi.remove_members_by_inbox_id(inbox_ids)

    async def add_members_by_identifiers(self, identifiers: list[Identifier]) -> None:
        """Add members to the group by identifier."""

        members = [_identifier_to_ffi(identifier) for identifier in identifiers]
        await self._ffi.add_members(members)

    async def remove_members_by_identifiers(self, identifiers: list[Identifier]) -> None:
        """Remove members from the group by identifier."""

        members = [_identifier_to_ffi(identifier) for identifier in identifiers]
        await self._ffi.remove_members(members)

    async def members(self) -> list[str]:
        """Return group member inbox IDs."""

        members = await self._ffi.list_members()
        return [member.inbox_id for member in members]

    async def add_admin(self, inbox_id: str) -> None:
        """Add an admin to the group."""

        await self._ffi.add_admin(inbox_id)

    async def remove_admin(self, inbox_id: str) -> None:
        """Remove an admin from the group."""

        await self._ffi.remove_admin(inbox_id)

    async def is_admin(self, inbox_id: str) -> bool:
        """Return True if inbox_id is an admin."""

        result = self._ffi.is_admin(inbox_id)
        if inspect.isawaitable(result):
            return await result
        return result

    async def is_super_admin(self, inbox_id: str) -> bool:
        """Return True if inbox_id is a super admin."""

        result = self._ffi.is_super_admin(inbox_id)
        if inspect.isawaitable(result):
            return await result
        return result

    @property
    def name(self) -> str | None:
        """Group name."""

        try:
            return self._ffi.group_name()
        except Exception:
            return None

    @property
    def description(self) -> str | None:
        """Group description."""

        try:
            return self._ffi.group_description()
        except Exception:
            return None

    @property
    def image_url(self) -> str | None:
        """Group image URL."""

        try:
            return self._ffi.group_image_url_square()
        except Exception:
            return None



"""XMTP agent SDK (unofficial)."""

from xmtp_agent.agent import Agent
from xmtp_agent.attachments import download_remote_attachment
from xmtp_agent.command_router import CommandRouter
from xmtp_agent.context import ClientContext, ConversationContext, MessageContext
from xmtp_agent.debug import get_installation_info, get_test_url, log_details
from xmtp_agent.filters import filter
from xmtp_agent.middleware import backoff_reconnect
from xmtp_agent.name_resolver import create_name_resolver
from xmtp_agent.recipient_resolver import ResolvedRecipient, resolve_recipient
from xmtp_agent.test_utils import (
    MockAsyncStream,
    create_mock_message,
    deserialize_message,
    flush_asyncio,
    record_messages,
    replay_messages,
    serialize_message,
)
from xmtp_agent.user import create_identifier, create_signer, create_user

__all__ = [
    "Agent",
    "backoff_reconnect",
    "ClientContext",
    "ConversationContext",
    "CommandRouter",
    "MessageContext",
    "MockAsyncStream",
    "ResolvedRecipient",
    "create_identifier",
    "create_mock_message",
    "create_name_resolver",
    "create_signer",
    "create_user",
    "deserialize_message",
    "download_remote_attachment",
    "filter",
    "flush_asyncio",
    "get_installation_info",
    "get_test_url",
    "log_details",
    "record_messages",
    "replay_messages",
    "resolve_recipient",
    "serialize_message",
]

__version__ = "0.1.5"



"""XMTP client entry point."""

from __future__ import annotations

import inspect
import os
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, TypeVar, cast

from xmtp_content_type_primitives import ContentCodec, ContentTypeId, EncodedContent

from xmtp.bindings import NativeBindings
from xmtp.conversations import Conversations
from xmtp.env import (
    apply_rust_log_from_options,
    load_client_options_from_env,
    load_signer_from_env,
)
from xmtp.errors import (
    ClientNotInitializedError,
    CodecNotFoundError,
    DatabaseOpenError,
    SignerUnavailableError,
)
from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.messages import DecodedMessage
from xmtp.preferences import Preferences
from xmtp.signers.base import Signer, SignerType
from xmtp.types import ClientOptions
from xmtp.utils import coerce_db_encryption_key

ContentT = TypeVar("ContentT")

_DB_ERROR_MARKERS = (
    "sqlcipher",
    "sqlite",
    "database",
    "db3",
    "file is encrypted",
    "not a database",
    "cipher",
    "malformed",
)


def _identifier_to_ffi(identifier: Identifier) -> NativeBindings.FfiIdentifier:
    kind = {
        IdentifierKind.ETHEREUM: NativeBindings.FfiIdentifierKind.ETHEREUM,
        IdentifierKind.PASSKEY: NativeBindings.FfiIdentifierKind.PASSKEY,
    }[identifier.kind]
    return NativeBindings.FfiIdentifier(identifier=identifier.value, identifier_kind=kind)


@dataclass(slots=True)
class _SendMessageOpts:
    should_push: bool


def _default_send_opts(
    should_push: bool = True,
) -> NativeBindings.FfiSendMessageOpts | _SendMessageOpts:
    try:
        return NativeBindings.FfiSendMessageOpts(should_push=should_push)
    except Exception:
        return _SendMessageOpts(should_push=should_push)


def _content_type_from_ffi(
    content_type: NativeBindings.FfiContentTypeId | None,
) -> ContentTypeId | None:
    if content_type is None:
        return None
    return ContentTypeId(
        authority_id=content_type.authority_id,
        type_id=content_type.type_id,
        version_major=content_type.version_major,
        version_minor=content_type.version_minor,
    )


def _encoded_from_ffi(encoded: NativeBindings.FfiEncodedContent) -> EncodedContent:
    content_type = _content_type_from_ffi(encoded.type_id)
    if content_type is None:
        raise ValueError("Missing content type in encoded content")
    return EncodedContent(
        type_id=content_type,
        parameters=encoded.parameters,
        fallback=encoded.fallback,
        compression=encoded.compression,
        content=encoded.content,
    )


def _looks_like_db_error(error: Exception) -> bool:
    message = str(error).lower()
    return any(marker in message for marker in _DB_ERROR_MARKERS)


class Client:
    """Main client for interacting with the XMTP network."""

    def __init__(self, options: ClientOptions | None = None) -> None:
        self._options = options or ClientOptions()
        self._signer: Signer | None = None
        self._identifier: Identifier | None = None
        self._client: NativeBindings.FfiXmtpClient | None = None
        self._conversations: Conversations | None = None
        self._preferences: Preferences | None = None
        self._codecs: dict[str, ContentCodec[object]] = {}
        self._register_default_codecs()
        if self._options.codecs:
            for codec in self._options.codecs:
                self.register_codec(codec)

    def _register_default_codecs(self) -> None:
        try:
            import xmtp_bindings  # noqa: F401
        except ImportError:
            return

        try:
            from xmtp_content_type_group_updated import GroupUpdatedCodec
            from xmtp_content_type_markdown import MarkdownCodec
            from xmtp_content_type_reaction import ReactionCodec
            from xmtp_content_type_read_receipt import ReadReceiptCodec
            from xmtp_content_type_remote_attachment import AttachmentCodec, RemoteAttachmentCodec
            from xmtp_content_type_reply import ReplyCodec
            from xmtp_content_type_text import TextCodec
            from xmtp_content_type_transaction_reference import TransactionReferenceCodec
            from xmtp_content_type_wallet_send_calls import WalletSendCallsCodec
        except ImportError:
            return

        for codec in (
            TextCodec(),
            MarkdownCodec(),
            ReactionCodec(),
            ReadReceiptCodec(),
            AttachmentCodec(),
            RemoteAttachmentCodec(),
            ReplyCodec(),
            GroupUpdatedCodec(),
            TransactionReferenceCodec(),
            WalletSendCallsCodec(),
        ):
            self.register_codec(cast(ContentCodec[object], codec))

    @classmethod
    async def create(cls, signer: Signer, options: ClientOptions | None = None) -> Client:
        """Create a client with a signer."""

        client = cls(options)
        client._signer = signer
        identifier = await signer.get_identifier()
        await client._init(identifier)
        if not client._options.disable_auto_register:
            await client.register()
        return client

    @classmethod
    async def create_from_env(cls, options: ClientOptions | None = None) -> Client:
        """Create a client using XMTP environment variables."""

        signer = load_signer_from_env()
        opts = load_client_options_from_env(options)
        return await cls.create(signer, opts)

    @classmethod
    async def from_env(cls, options: ClientOptions | None = None) -> Client:
        """Alias for create_from_env."""

        return await cls.create_from_env(options)

    @classmethod
    async def build(cls, identifier: Identifier, options: ClientOptions | None = None) -> Client:
        """Create a client with an identifier (no signer)."""

        client = cls(options)
        await client._init(identifier)
        return client

    async def _init(self, identifier: Identifier) -> None:
        if self._client is not None:
            return

        self._identifier = identifier
        options = self._options
        apply_rust_log_from_options(options)
        host = options.resolved_api_url()
        gateway_host = options.gateway_host
        is_secure = host.startswith("https")

        api = await NativeBindings.connect_to_backend(
            host,
            gateway_host,
            is_secure,
            None,
            options.app_version,
            None,
            None,
        )

        history_sync_url = options.resolved_history_sync_url()
        if history_sync_url:
            sync_api = await NativeBindings.connect_to_backend(
                history_sync_url,
                gateway_host,
                history_sync_url.startswith("https"),
                None,
                options.app_version,
                None,
                None,
            )
        else:
            sync_api = api
            history_sync_url = None

        ffi_identifier = _identifier_to_ffi(identifier)
        inbox_id = await NativeBindings.get_inbox_id_for_identifier(api, ffi_identifier)
        nonce = options.nonce if options.nonce is not None else 0
        if inbox_id is None:
            inbox_id = NativeBindings.generate_inbox_id(ffi_identifier, nonce)

        db_path_option = options.db_path
        db_path: str | None
        if db_path_option == "auto":
            db_path = os.path.join(os.getcwd(), f"xmtp-{options.env}-{inbox_id}.db3")
        elif callable(db_path_option):
            db_path = db_path_option(inbox_id)
        else:
            db_path = db_path_option

        encryption_key = coerce_db_encryption_key(options.db_encryption_key)

        device_sync_mode = (
            NativeBindings.FfiSyncWorkerMode.DISABLED
            if options.disable_device_sync
            else NativeBindings.FfiSyncWorkerMode.ENABLED
        )

        try:
            self._client = await NativeBindings.create_client(
                api,
                sync_api,
                db_path,
                encryption_key,
                inbox_id,
                ffi_identifier,
                nonce,
                None,
                history_sync_url,
                device_sync_mode,
                None,
                None,
            )
        except Exception as exc:
            if _looks_like_db_error(exc):
                raise DatabaseOpenError(db_path, str(exc)) from exc
            raise

        conversations = Conversations(self, self._client.conversations())
        self._conversations = conversations
        self._preferences = Preferences(self, conversations)

    @property
    def inbox_id(self) -> str | None:
        """Inbox identifier for the user."""

        if self._client is None:
            return None
        return self._client.inbox_id()

    @property
    def installation_id(self) -> bytes | None:
        """Installation identifier for the user."""

        if self._client is None:
            return None
        return self._client.installation_id()

    @property
    def account_identifier(self) -> Identifier | None:
        """Return the account identifier used to initialize the client."""

        return self._identifier

    @property
    def is_registered(self) -> bool:
        """Return True if the user is registered with XMTP."""

        if self._client is None:
            return False
        return self._client.signature_request() is None

    @property
    def conversations(self) -> Conversations:
        """Conversation manager for the client."""

        if self._conversations is None:
            raise ClientNotInitializedError()
        return self._conversations

    @property
    def options(self) -> ClientOptions:
        """Return the client options."""

        return self._options

    @property
    def preferences(self) -> Preferences:
        """Preferences manager for the client."""

        if self._preferences is None:
            raise ClientNotInitializedError()
        return self._preferences

    async def register(self) -> None:
        """Register the user on XMTP."""

        if self._client is None:
            raise ClientNotInitializedError()
        if self._signer is None:
            raise SignerUnavailableError()

        signature_request = self._client.signature_request()
        if signature_request is None:
            return

        signature_text_result = cast(Any, signature_request).signature_text()
        if inspect.isawaitable(signature_text_result):
            signature_text = await signature_text_result
        else:
            signature_text = signature_text_result
        signature = await self._signer.sign_message(signature_text.encode())

        if self._signer.type == SignerType.SCW:
            address = await self._signer.get_address()
            chain_id = await self._signer.get_chain_id()
            block_number = await self._signer.get_block_number()
            await signature_request.add_scw_signature(signature, address, chain_id, block_number)
        else:
            await signature_request.add_ecdsa_signature(signature)

        await self._client.register_identity(signature_request)

    async def can_message(self, identifiers: list[Identifier]) -> dict[str, bool]:
        """Return a map of identifiers to messageability."""

        if self._client is None:
            raise ClientNotInitializedError()

        ffi_identifiers = [_identifier_to_ffi(identifier) for identifier in identifiers]
        result = await self._client.can_message(ffi_identifiers)
        return {item.identifier: can for item, can in result.items()}

    async def get_inbox_id_by_identifier(self, identifier: Identifier) -> str | None:
        """Resolve an identifier to an inbox id."""

        if self._client is None:
            raise ClientNotInitializedError()
        return await self._client.find_inbox_id(_identifier_to_ffi(identifier))

    def register_codec(self, codec: ContentCodec[ContentT]) -> None:
        """Register a content codec for encoding/decoding."""

        self._codecs[str(codec.content_type)] = cast(ContentCodec[object], codec)

    def register_codecs(self, codecs: Sequence[ContentCodec[ContentT]]) -> None:
        """Register multiple content codecs."""

        for codec in codecs:
            self.register_codec(codec)

    def codec_for(self, content_type: ContentTypeId | str) -> ContentCodec[object] | None:
        """Return the codec for a content type, if registered."""

        return self._codecs.get(str(content_type))

    def encode_content(self, content: ContentT, content_type: ContentTypeId | str) -> bytes:
        """Encode content for sending."""

        codec = self.codec_for(content_type)
        if codec is None:
            raise CodecNotFoundError(str(content_type))
        encoded = cast(ContentCodec[ContentT], codec).encode(content, self)
        return encoded.content

    def prepare_for_send(
        self,
        content: ContentT,
        content_type: ContentTypeId | str,
    ) -> tuple[bytes, NativeBindings.FfiSendMessageOpts | _SendMessageOpts]:
        """Prepare content for sending with codec-derived send options."""

        codec = self.codec_for(content_type)
        if codec is None:
            raise CodecNotFoundError(str(content_type))
        encoded = cast(ContentCodec[ContentT], codec).encode(content, self)
        should_push = cast(ContentCodec[ContentT], codec).should_push(content)
        send_opts = _default_send_opts(should_push=should_push)
        return encoded.content, send_opts

    def _decode_message(self, message: NativeBindings.FfiMessage) -> DecodedMessage[object]:
        if self._client is None:
            raise ClientNotInitializedError()

        decoded = self._client.enriched_message(message.id)
        content = self._decode_ffi_content(decoded.content())
        sent_at = datetime.fromtimestamp(decoded.sent_at_ns() / 1_000_000_000, tz=timezone.utc)
        content_type = _content_type_from_ffi(decoded.content_type_id())
        content_type_id = str(content_type) if content_type is not None else None
        return DecodedMessage(
            id=decoded.id(),
            conversation_id=decoded.conversation_id(),
            sender_inbox_id=decoded.sender_inbox_id(),
            sent_at=sent_at,
            content=content,
            content_type_id=content_type_id,
        )

    def _decode_ffi_content(self, content: NativeBindings.FfiDecodedMessageContent) -> object:
        if content.is_TEXT():
            text_payload = cast("NativeBindings.FfiDecodedMessageContent.TEXT", content)
            return text_payload[0].content
        if content.is_MARKDOWN():
            markdown_payload = cast("NativeBindings.FfiDecodedMessageContent.MARKDOWN", content)
            return markdown_payload[0].content
        if content.is_REACTION():
            from xmtp_content_type_reaction import Reaction, ReactionAction, ReactionSchema

            reaction_payload = cast("NativeBindings.FfiDecodedMessageContent.REACTION", content)[0]
            action = (
                ReactionAction.ADDED
                if reaction_payload.action == NativeBindings.FfiReactionAction.ADDED
                else ReactionAction.REMOVED
            )
            schema_map = {
                NativeBindings.FfiReactionSchema.UNICODE: ReactionSchema.UNICODE,
                NativeBindings.FfiReactionSchema.SHORTCODE: ReactionSchema.SHORTCODE,
                NativeBindings.FfiReactionSchema.CUSTOM: ReactionSchema.CUSTOM,
            }
            return Reaction(
                reference=reaction_payload.reference,
                reference_inbox_id=reaction_payload.reference_inbox_id or None,
                action=action,
                content=reaction_payload.content,
                schema=schema_map[reaction_payload.schema],
            )
        if content.is_REPLY():
            from xmtp_content_type_reply import Reply

            reply_payload = cast("NativeBindings.FfiDecodedMessageContent.REPLY", content)[0]
            encoded = _encoded_from_ffi(reply_payload.content)
            codec = self.codec_for(encoded.type_id)
            nested_content = codec.decode(encoded, self) if codec is not None else encoded.content
            return Reply(
                reference=reply_payload.reference,
                reference_inbox_id=reply_payload.reference_inbox_id or None,
                content=nested_content,
                content_type=encoded.type_id,
            )
        if content.is_REMOTE_ATTACHMENT():
            from xmtp_content_type_remote_attachment import RemoteAttachment

            attachment = cast(
                "NativeBindings.FfiDecodedMessageContent.REMOTE_ATTACHMENT",
                content,
            )[0]
            return RemoteAttachment(
                url=attachment.url,
                content_digest=attachment.content_digest,
                secret=attachment.secret,
                salt=attachment.salt,
                nonce=attachment.nonce,
                scheme=attachment.scheme,
                content_length=attachment.content_length,
                filename=attachment.filename,
            )
        if content.is_READ_RECEIPT():
            return {}
        if content.is_TRANSACTION_REFERENCE():
            from xmtp_content_type_transaction_reference import (
                TransactionMetadata,
                TransactionReference,
            )

            transaction_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.TRANSACTION_REFERENCE",
                content,
            )[0]
            metadata = None
            if transaction_payload.metadata is not None:
                metadata = TransactionMetadata(
                    transaction_type=transaction_payload.metadata.transaction_type,
                    currency=transaction_payload.metadata.currency,
                    amount=transaction_payload.metadata.amount,
                    decimals=transaction_payload.metadata.decimals,
                    from_address=transaction_payload.metadata.from_address,
                    to_address=transaction_payload.metadata.to_address,
                )
            return TransactionReference(
                namespace=transaction_payload.namespace,
                network_id=transaction_payload.network_id,
                reference=transaction_payload.reference,
                metadata=metadata,
            )
        if content.is_WALLET_SEND_CALLS():
            from xmtp_content_type_wallet_send_calls import (
                WalletCall,
                WalletCallMetadata,
                WalletSendCalls,
            )

            wallet_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.WALLET_SEND_CALLS",
                content,
            )[0]
            calls: list[WalletCall] = []
            for call in wallet_payload.calls:
                call_metadata = None
                if call.metadata is not None:
                    call_metadata = WalletCallMetadata(
                        description=call.metadata.description,
                        transaction_type=call.metadata.transaction_type,
                        extra=call.metadata.extra,
                    )
                calls.append(
                    WalletCall(
                        to=call.to,
                        data=call.data,
                        value=call.value,
                        gas=call.gas,
                        metadata=call_metadata,
                    )
                )
            return WalletSendCalls(
                version=wallet_payload.version,
                chain_id=wallet_payload.chain_id,
                from_address=wallet_payload._from,
                calls=calls,
                capabilities=wallet_payload.capabilities,
            )
        if content.is_GROUP_UPDATED():
            group_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.GROUP_UPDATED",
                content,
            )
            return group_payload[0]
        if content.is_ATTACHMENT():
            from xmtp_content_type_remote_attachment import Attachment

            attachment = cast(
                "NativeBindings.FfiDecodedMessageContent.ATTACHMENT",
                content,
            )[0]
            return Attachment(
                filename=attachment.filename,
                mime_type=attachment.mime_type,
                data=attachment.content,
            )
        if content.is_ACTIONS():
            actions_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.ACTIONS",
                content,
            )
            return actions_payload[0]
        if content.is_INTENT():
            intent_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.INTENT",
                content,
            )
            return intent_payload[0]
        if content.is_LEAVE_REQUEST():
            leave_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.LEAVE_REQUEST",
                content,
            )
            return leave_payload[0]
        if content.is_CUSTOM():
            custom_payload = cast(
                "NativeBindings.FfiDecodedMessageContent.CUSTOM",
                content,
            )
            encoded = custom_payload[0]
            try:
                decoded = _encoded_from_ffi(encoded)
            except ValueError:
                return encoded
            codec = self.codec_for(decoded.type_id)
            if codec is None:
                return encoded
            return codec.decode(decoded, self)
        return content



"""Content type primitives for XMTP."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Protocol, TypeVar

TContent = TypeVar("TContent")


@dataclass(frozen=True, slots=True)
class ContentTypeId:
    """Content type identifier."""

    authority_id: str
    type_id: str
    version_major: int
    version_minor: int

    def __str__(self) -> str:
        return f"{self.authority_id}/{self.type_id}:{self.version_major}.{self.version_minor}"

    def same_as(self, other: ContentTypeId) -> bool:
        return (
            self.authority_id == other.authority_id
            and self.type_id == other.type_id
            and self.version_major == other.version_major
            and self.version_minor == other.version_minor
        )


@dataclass(slots=True)
class EncodedContent:
    """Encoded content payload."""

    type_id: ContentTypeId
    parameters: dict[str, str]
    content: bytes
    fallback: str | None = None
    compression: int | None = None


class CodecRegistry(Protocol):
    """Registry for content codecs."""

    def codec_for(self, content_type: ContentTypeId) -> ContentCodec[Any] | None: ...


class ContentCodec(Protocol[TContent]):
    """Protocol for content codecs."""

    @property
    def content_type(self) -> ContentTypeId: ...

    def encode(
        self, content: TContent, registry: CodecRegistry | None = None
    ) -> EncodedContent: ...

    def decode(
        self, content: EncodedContent, registry: CodecRegistry | None = None
    ) -> TContent: ...

    def fallback(self, content: TContent) -> str | None: ...

    def should_push(self, content: TContent) -> bool: ...


class BaseContentCodec(Generic[TContent]):
    """Base class for content codecs with default behaviors."""

    @property
    def content_type(self) -> ContentTypeId:
        raise NotImplementedError

    def encode(self, content: TContent, registry: CodecRegistry | None = None) -> EncodedContent:
        raise NotImplementedError

    def decode(self, content: EncodedContent, registry: CodecRegistry | None = None) -> TContent:
        raise NotImplementedError

    def fallback(self, content: TContent) -> str | None:
        return None

    def should_push(self, content: TContent) -> bool:
        return True


def content_types_are_equal(a: ContentTypeId, b: ContentTypeId) -> bool:
    """Return True if content type IDs are equal."""

    return a.same_as(b)


def content_type_to_string(content_type: ContentTypeId) -> str:
    """Convert content type ID to string form."""

    return str(content_type)


def content_type_from_string(content_type_string: str) -> ContentTypeId:
    """Parse content type ID from string form."""
    import re

    match = re.match(r"^([^/]+)/([^:]+):(\d+)\.(\d+)$", content_type_string)
    if not match:
        raise ValueError(
            f'Invalid content type string: "{content_type_string}". '
            'Expected format: "authorityId/typeId:majorVersion.minorVersion"'
        )
    authority_id, type_id, major_str, minor_str = match.groups()
    return ContentTypeId(
        authority_id=authority_id,
        type_id=type_id,
        version_major=int(major_str),
        version_minor=int(minor_str),
    )


__all__ = [
    "BaseContentCodec",
    "CodecRegistry",
    "ContentCodec",
    "ContentTypeId",
    "EncodedContent",
    "content_type_from_string",
    "content_type_to_string",
    "content_types_are_equal",
]



"""Shared XMTP type definitions."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Literal

from xmtp_content_type_primitives import ContentCodec

from xmtp.constants import API_URLS, HISTORY_SYNC_URLS

XmtpEnv = Literal["local", "dev", "production"]


class LogLevel(str, Enum):
    """Logging levels for libxmtp."""

    OFF = "off"
    ERROR = "error"
    WARN = "warn"
    INFO = "info"
    DEBUG = "debug"
    TRACE = "trace"


@dataclass(slots=True)
class ClientOptions:
    """Options for constructing an XMTP client.

    Attributes:
        env: Network environment to target.
        api_url: Optional override for the XMTP API URL.
        history_sync_url: Optional override for history sync URL.
        disable_history_sync: Disable history sync and use the primary API for identity calls.
            Defaults to True.
        gateway_host: Optional gateway host for d14n.
        db_path: Optional database path (string, None, or function).
        db_encryption_key: Optional database encryption key bytes or hex string.
        structured_logging: Enable structured JSON logging.
        logging_level: Logging level for libxmtp.
        rust_log: Override RUST_LOG for native bindings (defaults to "off" if unset).
        disable_auto_register: Skip automatic registration.
        disable_device_sync: Disable device sync worker.
        app_version: Optional app version string to pass to the backend.
        debug_events_enabled: Enable debug events tracking.
        nonce: Nonce for generating inbox IDs when needed.
        codecs: Optional content codecs to register.
    """

    env: XmtpEnv = "dev"
    api_url: str | None = None
    history_sync_url: str | None = None
    disable_history_sync: bool = True
    gateway_host: str | None = None
    db_path: str | None | Literal["auto"] | Callable[[str], str] = "auto"
    db_encryption_key: bytes | str | None = None
    structured_logging: bool = False
    logging_level: LogLevel | None = None
    rust_log: str | None = None
    disable_auto_register: bool = False
    disable_device_sync: bool = False
    app_version: str | None = None
    debug_events_enabled: bool = False
    nonce: int | None = None
    codecs: Sequence[ContentCodec[object]] | None = None

    def resolved_api_url(self) -> str:
        """Return the API URL based on env and overrides."""

        if self.api_url:
            return self.api_url
        return API_URLS[self.env]

    def resolved_history_sync_url(self) -> str | None:
        """Return the history sync URL based on env and overrides."""

        if self.disable_history_sync:
            return None
        if self.history_sync_url is not None:
            return self.history_sync_url
        return HISTORY_SYNC_URLS[self.env]



"""Identifier types for XMTP."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


class IdentifierKind(str, Enum):
    """Supported identifier kinds."""

    ETHEREUM = "ethereum"
    PASSKEY = "passkey"


@dataclass(frozen=True, slots=True)
class Identifier:
    """Identifier for XMTP inbox lookups."""

    kind: IdentifierKind
    value: str



"""User preferences management."""

from __future__ import annotations

from typing import TYPE_CHECKING

from xmtp.bindings import NativeBindings
from xmtp.errors import ClientNotInitializedError

if TYPE_CHECKING:
    from xmtp.client import Client
    from xmtp.conversations import Conversations


class Preferences:
    """Manage user preferences for the XMTP client."""

    def __init__(self, client: Client, conversations: Conversations) -> None:
        self._client = client
        self._conversations = conversations

    def _ensure(self) -> NativeBindings.FfiXmtpClient:
        if self._client._client is None:
            raise ClientNotInitializedError()
        return self._client._client

    async def refresh(self) -> None:
        """Refresh preferences from the network."""

        await self._ensure().sync_preferences()

    async def inbox_state(self, refresh_from_network: bool = False) -> NativeBindings.FfiInboxState:
        """Return the current inbox state."""

        return await self._ensure().inbox_state(refresh_from_network)

    async def get_latest_inbox_state(self, inbox_id: str) -> NativeBindings.FfiInboxState:
        """Return the latest inbox state for a specific inbox."""

        return await self._ensure().get_latest_inbox_state(inbox_id)

    async def inbox_state_from_inbox_ids(
        self, inbox_ids: list[str], refresh_from_network: bool = False
    ) -> list[NativeBindings.FfiInboxState]:
        """Return inbox state for the specified inbox IDs."""

        return await self._ensure().addresses_from_inbox_id(refresh_from_network, inbox_ids)

    async def set_consent_states(self, records: list[NativeBindings.FfiConsent]) -> None:
        """Update consent states for multiple records."""

        await self._ensure().set_consent_states(records)

    async def get_consent_state(
        self, entity_type: NativeBindings.FfiConsentEntityType, entity: str
    ) -> NativeBindings.FfiConsentState:
        """Get consent state for a specific entity."""

        return await self._ensure().get_consent_state(entity_type, entity)



"""XMTP network constants."""

from __future__ import annotations

from typing import Final

API_URLS: Final[dict[str, str]] = {
    "local": "http://localhost:5556",
    "dev": "https://grpc.dev.xmtp.network:443",
    "production": "https://grpc.production.xmtp.network:443",
}

HISTORY_SYNC_URLS: Final[dict[str, str]] = {
    "local": "http://localhost:5558",
    "dev": "https://message-history.dev.ephemera.network",
    "production": "https://message-history.production.ephemera.network",
}



"""Debug helpers for agent development."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from xmtp import Client

if TYPE_CHECKING:
    from xmtp_agent.agent import Agent


def get_test_url(client: Client) -> str:
    """Return a URL to test the agent on xmtp.chat."""

    address = client.account_identifier.value if client.account_identifier else None
    env = client.options.env
    return f"http://xmtp.chat/{env}/dm/{address}"


@dataclass(slots=True)
class InstallationInfo:
    """Installation details for an XMTP client."""

    total_installations: int
    installation_id: str | None
    most_recent_installation_id: str | None
    is_most_recent: bool


def _installation_sort_key(installation: object) -> int:
    timestamp = getattr(installation, "client_timestamp_ns", None)
    if timestamp is None:
        return 0
    return int(timestamp)


async def get_installation_info(client: Client) -> InstallationInfo:
    """Return information about the client's installation state."""

    inbox_id = client.inbox_id
    installation_id = client.installation_id
    if inbox_id is None or installation_id is None:
        return InstallationInfo(
            total_installations=0,
            installation_id=None,
            most_recent_installation_id=None,
            is_most_recent=False,
        )

    inbox_state = await client.preferences.inbox_state(refresh_from_network=True)
    installations = list(getattr(inbox_state, "installations", []) or [])
    installations.sort(key=_installation_sort_key, reverse=True)

    installation_id_hex = installation_id.hex()
    most_recent = installations[0] if installations else None
    most_recent_hex = most_recent.id.hex() if most_recent else None
    is_most_recent = most_recent_hex == installation_id_hex if most_recent_hex else False

    return InstallationInfo(
        total_installations=len(installations),
        installation_id=installation_id_hex,
        most_recent_installation_id=most_recent_hex,
        is_most_recent=is_most_recent,
    )


async def log_details(agent: Agent) -> None:
    """Log basic agent details for debugging."""

    client = agent.client
    inbox_id = client.inbox_id
    installation_id = client.installation_id
    installation_id_display = installation_id.hex() if installation_id else None
    address = client.account_identifier.value if client.account_identifier else None
    env = client.options.env

    conversations = await client.conversations.list()
    inbox_state = await client.preferences.inbox_state()

    print("XMTP Agent Details")
    print(f"- Inbox ID: {inbox_id}")
    print(f"- Installation ID: {installation_id_display}")
    print(f"- Address: {address}")
    print(f"- Conversations: {len(conversations)}")
    print(f"- Installations: {len(inbox_state.installations)}")
    print(f"- Environment: {env}")
    print(f"- Test URL: {get_test_url(client)}")


__all__ = ["get_test_url", "get_installation_info", "log_details", "InstallationInfo"]



"""Reply content type for XMTP."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, cast

from xmtp_content_type_primitives import (
    CodecRegistry,
    ContentCodec,
    ContentTypeId,
    EncodedContent,
)


class _FfiContentTypeId(Protocol):
    def __init__(
        self,
        *,
        authority_id: str,
        type_id: str,
        version_major: int,
        version_minor: int,
    ) -> None: ...

    authority_id: str
    type_id: str
    version_major: int
    version_minor: int


class _FfiEncodedContent(Protocol):
    def __init__(
        self,
        *,
        type_id: _FfiContentTypeId,
        parameters: dict[str, str],
        fallback: str | None,
        compression: int | None,
        content: bytes,
    ) -> None: ...

    type_id: _FfiContentTypeId | None
    parameters: dict[str, str]
    fallback: str | None
    compression: int | None
    content: bytes


class _FfiReply(Protocol):
    def __init__(
        self,
        *,
        reference: str,
        reference_inbox_id: str | None,
        content: _FfiEncodedContent,
    ) -> None: ...

    reference: str
    reference_inbox_id: str | None
    content: _FfiEncodedContent


class _Bindings(Protocol):
    FfiContentTypeId: type[_FfiContentTypeId]
    FfiEncodedContent: type[_FfiEncodedContent]
    FfiReply: type[_FfiReply]

    def encode_reply(self, payload: _FfiReply) -> bytes: ...

    def decode_reply(self, data: bytes) -> _FfiReply: ...


def _bindings() -> _Bindings:  # pragma: no cover - requires native bindings
    from xmtp_bindings import xmtpv3  # pragma: no cover - requires native bindings

    return cast(_Bindings, xmtpv3)  # pragma: no cover - requires native bindings


ContentTypeReply = ContentTypeId(
    authority_id="xmtp.org",
    type_id="reply",
    version_major=1,
    version_minor=0,
)


@dataclass(slots=True)
class Reply:
    """Reply content wrapper."""

    reference: str
    reference_inbox_id: str | None
    content: Any
    content_type: ContentTypeId


def _content_type_to_ffi(content_type: ContentTypeId) -> _FfiContentTypeId:
    return _bindings().FfiContentTypeId(
        authority_id=content_type.authority_id,
        type_id=content_type.type_id,
        version_major=content_type.version_major,
        version_minor=content_type.version_minor,
    )


def _content_type_from_ffi(ffi: _FfiContentTypeId | None) -> ContentTypeId:
    if ffi is None:
        raise ValueError("Missing content type in encoded reply")
    return ContentTypeId(
        authority_id=ffi.authority_id,
        type_id=ffi.type_id,
        version_major=ffi.version_major,
        version_minor=ffi.version_minor,
    )


def _encoded_to_ffi(encoded: EncodedContent) -> _FfiEncodedContent:
    return _bindings().FfiEncodedContent(
        type_id=_content_type_to_ffi(encoded.type_id),
        parameters=encoded.parameters,
        fallback=encoded.fallback,
        compression=encoded.compression,
        content=encoded.content,
    )


def _encoded_from_ffi(encoded: _FfiEncodedContent) -> EncodedContent:
    return EncodedContent(
        type_id=_content_type_from_ffi(encoded.type_id),
        parameters=encoded.parameters,
        fallback=encoded.fallback,
        compression=encoded.compression,
        content=encoded.content,
    )


class ReplyCodec(ContentCodec[Reply]):
    """Codec for reply messages."""

    @property
    def content_type(self) -> ContentTypeId:
        return ContentTypeReply

    def encode(self, content: Reply, registry: CodecRegistry | None = None) -> EncodedContent:
        if registry is None:
            raise ValueError("Codec registry required to encode replies")

        codec = registry.codec_for(content.content_type)
        if codec is None:
            raise ValueError(f"Missing codec for content type {content.content_type}")

        encoded_content = codec.encode(content.content, registry)
        reply_payload = _bindings().FfiReply(
            reference=content.reference,
            reference_inbox_id=content.reference_inbox_id,
            content=_encoded_to_ffi(encoded_content),
        )
        reply_bytes = _bindings().encode_reply(reply_payload)
        parameters = {
            "contentType": str(content.content_type),
            "reference": content.reference,
        }
        if content.reference_inbox_id:
            parameters["referenceInboxId"] = content.reference_inbox_id
        return EncodedContent(
            type_id=self.content_type,
            parameters=parameters,
            content=reply_bytes,
        )

    def decode(self, content: EncodedContent, registry: CodecRegistry | None = None) -> Reply:
        if registry is None:
            raise ValueError("Codec registry required to decode replies")

        reply_payload = _bindings().decode_reply(content.content)
        decoded_content = _encoded_from_ffi(reply_payload.content)
        nested_content_type = decoded_content.type_id
        codec = registry.codec_for(nested_content_type)
        if codec is None:
            raise ValueError(f"Missing codec for content type {nested_content_type}")

        return Reply(
            reference=reply_payload.reference,
            reference_inbox_id=reply_payload.reference_inbox_id or None,
            content=codec.decode(decoded_content, registry),
            content_type=nested_content_type,
        )

    def fallback(self, content: Reply) -> str | None:
        if isinstance(content.content, str):
            return f'Replied with "{content.content}" to an earlier message'
        return "Replied to an earlier message"

    def should_push(self, content: Reply) -> bool:
        return True


__all__ = ["ContentTypeReply", "Reply", "ReplyCodec"]



"""Signer protocol for XMTP."""

from __future__ import annotations

from enum import Enum
from typing import Protocol

from xmtp.identifiers import Identifier


class SignerType(str, Enum):
    """Supported signer types."""

    EOA = "EOA"
    SCW = "SCW"


class Signer(Protocol):
    """Signer interface for XMTP clients."""

    type: SignerType

    async def get_address(self) -> str:
        """Return the wallet address for this signer."""

        ...

    async def get_identifier(self) -> Identifier:
        """Return the XMTP identifier for this signer."""

        ...

    async def sign_message(self, message: bytes) -> bytes:
        """Sign a message and return the signature bytes."""

        ...

    async def get_chain_id(self) -> int:
        """Return the chain ID (SCW only)."""

        ...

    async def get_block_number(self) -> int | None:
        """Return the block number for SCW signatures, if needed."""

        ...



"""Externally owned account signer."""

from __future__ import annotations

from typing import cast

from eth_account import Account
from eth_account.messages import encode_defunct

from xmtp.identifiers import Identifier, IdentifierKind
from xmtp.signers.base import Signer, SignerType


class EoaSigner(Signer):
    """EOA signer created from a private key."""

    def __init__(self, private_key: str) -> None:
        self._private_key = private_key
        self._account = Account.from_key(private_key)
        self.type = SignerType.EOA

    async def get_address(self) -> str:
        """Return the wallet address for this signer."""

        return cast(str, self._account.address)

    async def get_identifier(self) -> Identifier:
        """Return the XMTP identifier for this signer."""

        return Identifier(kind=IdentifierKind.ETHEREUM, value=self._account.address)

    async def sign_message(self, message: bytes) -> bytes:
        """Sign a message and return the signature bytes."""

        signable = encode_defunct(message)
        signed = self._account.sign_message(signable)
        return cast(bytes, signed.signature)

    async def get_chain_id(self) -> int:
        """Return the chain ID (not applicable for EOA)."""

        raise ValueError("EOA signer does not support chain_id")

    async def get_block_number(self) -> int | None:
        """Return block number (not applicable for EOA)."""

        return None


def create_signer(private_key: str) -> EoaSigner:
    """Create an EOA signer from a private key string."""

    return EoaSigner(private_key)



"""Message models for XMTP."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Generic, TypeVar

ContentT = TypeVar("ContentT")


@dataclass(slots=True)
class DecodedMessage(Generic[ContentT]):
    """Decoded message representation."""

    id: bytes
    conversation_id: bytes
    sender_inbox_id: str
    sent_at: datetime
    content: ContentT
    content_type_id: str | None = None
